"""
Blast output conversion in GFF requires a BLAST+ tabular format which can be
obtained by using the `--outfmt 6` option with the default columns, as
specified in :func:`mgkit.io.blast.parse_blast_tab`. The script can get data
from the standard in and ouputs GFF lines on the standard output by default.

Uniprot
*******

The Function :func:`mgkit.io.blast.parse_uniprot_blast` is used, which filters
BLAST hits based on bitscore and adds by default a *db* attribute to the
annotation with the value `UNIPROT-SP`, indicating that the SwissProt db is
used and a *dbq* attribute with the value 10. The feature type used in the GFF
is CDS.

.. blockdiag::

    {
        "BLAST+" [color = "#377eb8" , textcolor = 'white',
            shape = flowchart.input];
        "parse_uniprot_blast" [color = "#e41a1c" , textcolor = 'white',
            width=200, fontsize=16];
        "GFF" [color = "#4daf4a" , textcolor = 'white'];
        "BLAST+"  -> "parse_uniprot_blast" -> GFF;
    }

BlastDB
*******

If a BlastDB, such as *nt* or *nr* was used, the **blastdb** command offers
some quick defaults to parse BLAST results.

It now includes options to control the way the sequence header are formatted.
Options to change the separator used, as well as the column used as *gene_id*.
This was added because at the moment the GI identifier (the second column in
the header) is used, but it's being phased out in favour of the embl/gb/dbj
(right now the fourth column in the header). This should easy the transition to
the new format and makes it easier to adapt an older pipeline/blastdb to newer
files (like the ID to TAXA files).

The header from the a *ncbi-nt* header looks like this::

    gi|160361034|gb|CP000884.1

This is the default output accepted by the *blastdb* command. The fields are
separated by **|** (pipe) and the GI is used (--gene-index 1, since internally
the string is split by the separator and the second element is take - lists
indices are 0-based in Python). This output uses the following options::

    --header-sep '|' --gene-index 1

Notice the single quotes to pass the pipe symbol, since *bash* would interpret
it as pipeing to the next coommand otherwise. This is the default.

In case, for the same header, we want to use the *gb* identifier, the only
option to be specified is::

    --gene-index 3

This will get the fourth element of the header (since we're splitting by pipe).

As in the *uniprot* command, the *gene_id* can be set to use the whole header,
using the *-n* option. Useful in case the *BLAST* db that was used was custom
made. While pipe is used in major databases, it was made the default, by if the
db used has different conventions the separator can be changed. There's also
the options of later changing the *gene_id* in the output GFF if necessary.

Changes
*******

.. versionchanged:: 0.2.6
    added *-r* option to *blastdb*

.. versionchanged:: 0.2.5
    added more options to give user control to the *blastdb* command

.. versionadded:: 0.2.3
    added *--fasta-file* option, added more data from a blsat hit

.. versionadded:: 0.2.2
    added *blastdb* command

.. versionchanged:: 0.2.1

    added *-ft* option

.. versionchanged:: 0.1.13

* added *-n* parameter to *uniprot* command
* added *-k* option to *uniprot* command

.. versionadded:: 0.1.12

"""
import sys
import argparse
import logging
from .. import logger
from . import utils
from ..io import blast, fasta


LOG = logging.getLogger(__name__)


def parse_attr_arg(value):
    values = value.split(':')
    if len(values) != 2:
        raise argparse.ArgumentTypeError(
            "Wrong key/value format, must be 'key:value' 'key:value'"
        )

    return values[0], values[1]


def set_common_options(parser):
    parser.add_argument(
        '-dbq',
        '--db-quality',
        action='store',
        type=int,
        help='Quality of the DB used',
        default=10
    )
    parser.add_argument(
        '-b',
        '--bitscore',
        action='store',
        type=float,
        help='Minimum bitscore to keep the annotation',
        default=0.0
    )
    parser.add_argument(
        '-k',
        '--attr-value',
        action='append',
        type=parse_attr_arg,
        help='''Additional attribute and value to add to each annotation,
                in the form attr:value''',
        default=None
    )
    parser.add_argument(
        '-ft',
        '--feat-type',
        action='store',
        default='CDS',
        help='Feature type to use in the GFF'
    )
    parser.add_argument(
        '-a',
        '--fasta-file',
        type=argparse.FileType('r'),
        default=None,
        help="""
        Fasta file with nucleotide sequences, used to calculate the frame, if
        not used, the frame on the '-' strand will always be 0
        """
    )
    parser.add_argument(
        'input_file',
        nargs='?',
        type=argparse.FileType('r'),
        default='-',
        help='BLAST+ output file in tabular format, defaults to stdin'
    )
    parser.add_argument(
        'output_file',
        nargs='?',
        type=argparse.FileType('w'),
        default=sys.stdout,
        help='Output GFF file, defaults to stdout'
    )


def set_uniprot_parser(parser):
    parser.add_argument(
        '-db',
        '--db-used',
        action='store',
        type=str,
        default='UNIPROT-SP',
        help='Uniprot database used with BLAST'
    )
    parser.add_argument(
        '-n',
        '--no-split',
        action='store_true',
        default=False,
        help='''if used, the script assumes that the sequence header contains
                only the gene id'''
    )

    parser.set_defaults(func=convert_from_uniprot)


def set_blastdb_parser(parser):
    """
    .. versionadded:: 0.2.2
    """
    parser.add_argument(
        '-db',
        '--db-used',
        action='store',
        type=str,
        default='NCBI-NT',
        help='blastdb used'
    )
    parser.add_argument(
        '-n',
        '--no-split',
        action='store_true',
        default=False,
        help='''if used, the script assumes that the sequence header will be
                used as gene_id'''
    )
    parser.add_argument(
        '-s',
        '--header-sep',
        action='store',
        default='|',
        help="""The separator for the header, defaults to '|' (pipe)"""
    )
    parser.add_argument(
        '-i',
        '--gene-index',
        action='store',
        default=1,
        type=int,
        help="""Which of the header columns (0-based) to use as gene_id
                (defaults to 1 - the second column)"""
    )
    parser.add_argument(
        '-r',
        '--remove-version',
        action='store_true',
        default=False,
        help='''if used, the script removes the *version* information from the
                gene_id'''
    )

    parser.set_defaults(func=convert_from_blastdb)


def load_fasta_file(file_name):
    if file_name is None:
        return None

    return dict(
        (name, len(seq))
        for name, seq in fasta.load_fasta(file_name)
    )


def convert_from_blastdb(options):
    """
    .. versionadded:: 0.2.2
    """

    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    seqs = load_fasta_file(options.fasta_file)

    if options.no_split:
        name_func = lambda x: x
    else:
        if options.remove_version:
            name_func = lambda x: x.split(options.header_sep)[options.gene_index].split('.')[0]
        else:
            name_func = lambda x: x.split(options.header_sep)[options.gene_index]

    iterator = blast.parse_uniprot_blast(
        options.input_file,
        bitscore=options.bitscore,
        db=options.db_used,
        dbq=options.db_quality,
        name_func=name_func,
        feat_type=options.feat_type,
        seq_lengths=seqs
    )

    for annotation in iterator:
        if options.attr_value is not None:
            for key, value in options.attr_value:
                annotation.set_attr(key, value)

        annotation.to_file(options.output_file)


def convert_from_uniprot(options):

    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    seqs = load_fasta_file(options.fasta_file)

    if options.no_split is True:
        name_func = lambda x: x
    else:
        name_func = None

    iterator = blast.parse_uniprot_blast(
        options.input_file,
        bitscore=options.bitscore,
        db=options.db_used,
        dbq=options.db_quality,
        name_func=name_func,
        feat_type=options.feat_type,
        seq_lengths=seqs
    )

    for annotation in iterator:
        if options.attr_value is not None:
            for key, value in options.attr_value:
                annotation.set_attr(key, value)

        annotation.to_file(options.output_file)


def set_parser():
    """
    .. versionchanged:: 0.2.2
        added *blastdb* command

    Sets command line arguments parser
    """
    parser = argparse.ArgumentParser(
        description='Convert BLAST output to a GFF file',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    subparsers = parser.add_subparsers()
    parser_u = subparsers.add_parser(
        'uniprot',
        help='Blast results from a Uniprot database, by default SwissProt'
    )

    set_uniprot_parser(parser_u)
    set_common_options(parser_u)
    utils.add_basic_options(parser_u, manual=__doc__)

    parser_blastdb = subparsers.add_parser(
        'blastdb',
        help='Blast results from a NCBI database, like *nt*'
    )

    set_blastdb_parser(parser_blastdb)
    set_common_options(parser_blastdb)
    utils.add_basic_options(parser_blastdb, manual=__doc__)

    utils.add_basic_options(parser, manual=__doc__)

    return parser


def main():
    "Main function"

    options = set_parser().parse_args()

    logger.config_log(options.verbose)
    options.func(options)



"""
Interleave/deinterleave paired-end fastq files.

.. versionchanged:: 0.3.0
    added *convert* command to FASTA

"""
from __future__ import division

import random
import argparse
import logging
import mgkit
import HTSeq
from itertools import izip
from mgkit.io.fastq import choose_header_type
from mgkit.io.fastq import write_fastq_sequence
from mgkit.io import fasta
from . import utils

LOG = logging.getLogger(__name__)


def set_parser():
    """
    Sets command line arguments parser
    """
    parser = argparse.ArgumentParser(
        description='Interleave/Deinterleave fastq files',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    utils.add_basic_options(parser, manual=__doc__)
    # Deinterleave
    subparsers = parser.add_subparsers()
    parser_d = subparsers.add_parser('di', help='Deinterleave sequences')
    parser_d.add_argument(
        'input_file',
        type=argparse.FileType('r'),
        action='store',
        help="File with interleaved sequences"
    )
    parser_d.add_argument(
        'mate1_file',
        type=argparse.FileType('w'),
        action='store',
        help="File with mate 1 reads"
    )
    parser_d.add_argument(
        'mate2_file',
        type=argparse.FileType('w'),
        action='store',
        help="File with mate 2 reads"
    )
    parser_d.add_argument(
        '-s',
        '--strip',
        action='store_true',
        default=False,
        help="Strip additional info"
    )
    parser_d.set_defaults(func=deinterleave)
    # Interleave
    parser_i = subparsers.add_parser('il', help='Interleave sequences')
    parser_i.add_argument(
        'mate1_file',
        type=argparse.FileType('r'),
        action='store',
        help="File with mate 1 reads"
    )
    parser_i.add_argument(
        'mate2_file',
        type=argparse.FileType('r'),
        action='store',
        help="File with mate 1 reads"
    )
    parser_i.add_argument(
        'output_file',
        default='output-il.fq',
        type=argparse.FileType('w'),
        action='store',
        help='Output file name'
    )
    parser_i.set_defaults(func=interleave)
    # Sort
    parser_s = subparsers.add_parser('sort', help='Sort paired-end sequences')
    parser_s.add_argument(
        'mate1_input',
        type=argparse.FileType('r'),
        action='store',
        help="File with mate 1 reads"
    )
    parser_s.add_argument(
        'mate2_input',
        type=argparse.FileType('r'),
        action='store',
        help="File with mate 1 reads"
    )
    parser_s.add_argument(
        'mate1_output',
        type=argparse.FileType('w'),
        action='store',
        help="Output file with mate 1 reads"
    )
    parser_s.add_argument(
        'mate2_output',
        type=argparse.FileType('w'),
        action='store',
        help="Output file with mate 2 reads"
    )
    parser_s.set_defaults(func=sort)
    # Randomise
    parser_r = subparsers.add_parser('rand', help='Randomise sequences')
    parser_r.add_argument(
        'input_file',
        type=argparse.FileType('r'),
        action='store',
        help="File with sequences"
    )
    parser_r.add_argument(
        'output_file',
        type=argparse.FileType('w'),
        action='store',
        help="Output file"
    )
    parser_r.add_argument(
        '-b',
        '--buffer-size',
        type=int,
        action='store',
        default=None,
        help="Maximum number of sequences to keep in memory (default: all)"
    )
    parser_r.set_defaults(func=randomise)

    # Convert to fasta
    parser_convert = subparsers.add_parser('convert', help='Convert to FASTA')
    parser_convert.add_argument(
        'input_file',
        type=argparse.FileType('r'),
        action='store',
        help="Input FastQ file"
    )
    parser_convert.add_argument(
        'output_file',
        type=argparse.FileType('w'),
        action='store',
        help="Output FASTA file"
    )
    parser_convert.set_defaults(func=convert_command)

    return parser


def convert_command(options):
    LOG.info(
        "Reading from file (%s)",
        options.input_file.name
    )
    file_handle = HTSeq.FastqReader(options.input_file)

    for sequence in file_handle:
        fasta.write_fasta_sequence(
            options.output_file,
            sequence.name,
            sequence.seq
        )


def report_counts(count, wcount, counter=None):
    "Logs the status"
    if (counter is None) or (counter % 100000 == 0):
        LOG.info(
            "Read %-7d sequences, Wrote %-7d (%.2f%%)",
            count,
            wcount,
            wcount / count * 100
        )


def randomise(options):
    "Used for testing"
    LOG.info(
        "Reading from file (%s)",
        options.input_file.name
    )

    file_handle = HTSeq.FastqReader(options.input_file)
    random.seed(random.randint(0, 1000000))

    count = 0
    wcount = 0

    seqs = []

    for sequence in file_handle:
        count += 1
        report_counts(count, wcount, count)
        seqs.append((sequence.name, sequence.seq, sequence.qualstr))

        if (options.buffer_size is None) or (len(seqs) < options.buffer_size):
            continue

        index = random.randint(0, len(seqs) - 1)

        write_fastq_sequence(options.output_file, *seqs[index])

        del seqs[index]

        wcount += 1

    # Shuffle the list
    random.shuffle(seqs)

    # Flush the rest of the sequences to disk
    for sequence in seqs:
        write_fastq_sequence(options.output_file, *sequence)
        wcount += 1
        report_counts(count, wcount, wcount)

    report_counts(count, wcount, None)


def sort(options):
    "Sort two fastq files"
    LOG.info(
        "Reading from mate1 (%s) and mate2 (%s)",
        options.mate1_input.name,
        options.mate2_input.name
    )

    regex = None
    simple_header = False

    mate1 = {}
    mate2 = {}

    count = 0
    wcount = 0

    for sequence1, sequence2 in izip(HTSeq.FastqReader(options.mate1_input),
                                     HTSeq.FastqReader(options.mate2_input)):

        count += 1

        if (regex is None) and (not simple_header):
            regex = choose_header_type(sequence1.name)
            if regex is None:
                simple_header = True
                LOG.info("Using a simple header structure")

        if simple_header:
            key1 = sequence1.name[:-1]
            key2 = sequence2.name[:-1]
        else:
            match1 = regex.search(sequence1.name)
            match2 = regex.search(sequence2.name)

            key1 = (
                match1.group('lane'),
                match1.group('tile'),
                match1.group('xcoord'),
                match1.group('ycoord')
            )
            key2 = (
                match2.group('lane'),
                match2.group('tile'),
                match2.group('xcoord'),
                match2.group('ycoord')
            )

        seq1 = (sequence1.name, sequence1.seq, sequence1.qualstr)
        seq2 = (sequence2.name, sequence2.seq, sequence2.qualstr)

        if key1 == key2:
            # if the 2
            write_fastq_sequence(options.mate1_output, *seq1)
            write_fastq_sequence(options.mate2_output, *seq2)
            wcount += 1
            report_counts(count, wcount, count)
            continue

        mate1[key1] = seq1
        mate2[key2] = seq2

        if key1 in mate2:
            write_fastq_sequence(options.mate1_output, *mate1[key1])
            write_fastq_sequence(options.mate2_output, *mate2[key1])
            del mate1[key1]
            del mate2[key1]
            wcount += 1
        if key2 in mate1:
            write_fastq_sequence(options.mate1_output, *mate1[key2])
            write_fastq_sequence(options.mate2_output, *mate2[key2])
            del mate1[key2]
            del mate2[key2]
            wcount += 1

        report_counts(count, wcount, count)

    report_counts(count, wcount, None)


def deinterleave(options):
    "Deinterleave a fastq file"

    LOG.info("Reading from file %s", options.input_file.name)

    fastq_in = HTSeq.FastqReader(options.input_file)

    regex = None
    simple_header = False

    mate1 = {}
    mate2 = {}

    count = 0
    wcount = 0

    for sequence in fastq_in:

        count += 1

        if (regex is None) and (not simple_header):
            regex = choose_header_type(sequence.name)
            if regex is None:
                LOG.info("Using a simple header structure")
                simple_header = True

        if simple_header:
            key = sequence.name[:-1]
            mate = int(sequence.name[-1])
        else:
            match = regex.search(sequence.name)
            key = (
                match.group('lane'),
                match.group('tile'),
                match.group('xcoord'),
                match.group('ycoord')
            )
            mate = int(match.group('mate'))

        if options.strip:
            sequence_name = sequence.name.split('\t')[0]
        else:
            sequence_name = sequence.name

        if mate == 1:
            mate1[key] = (sequence_name, sequence.seq, sequence.qualstr)
        else:
            mate2[key] = (sequence_name, sequence.seq, sequence.qualstr)

        try:
            # if sequence header in both
            seq1 = mate1[key]
            seq2 = mate2[key]
            write_fastq_sequence(options.mate1_file, *seq1)
            write_fastq_sequence(options.mate2_file, *seq2)
            wcount += 2
            del mate1[key]
            del mate2[key]
        except KeyError:
            pass

        report_counts(count, wcount, count)

    report_counts(count, wcount, None)


def interleave(options):
    "Interleave two fastq files"

    LOG.info(
        "Reading from mate1 (%s) and mate2 (%s)",
        options.mate1_file.name,
        options.mate2_file.name
    )

    regex = None
    simple_header = False

    mate1 = {}
    mate2 = {}

    count = 0
    wcount = 0

    for sequence1, sequence2 in izip(HTSeq.FastqReader(options.mate1_file),
                                     HTSeq.FastqReader(options.mate2_file)):

        count += 1

        if (regex is None) and (not simple_header):
            regex = choose_header_type(sequence1.name)
            if regex is None:
                simple_header = True
                LOG.info("Using a simple header structure")

        if simple_header:
            key1 = sequence1.name[:-1]
            key2 = sequence2.name[:-1]
        else:
            match1 = regex.search(sequence1.name)
            match2 = regex.search(sequence2.name)

            key1 = (
                match1.group('lane'),
                match1.group('tile'),
                match1.group('xcoord'),
                match1.group('ycoord')
            )
            key2 = (
                match2.group('lane'),
                match2.group('tile'),
                match2.group('xcoord'),
                match2.group('ycoord')
            )

        seq1 = (sequence1.name, sequence1.seq, sequence1.qualstr)
        seq2 = (sequence2.name, sequence2.seq, sequence2.qualstr)

        if key1 == key2:
            # if the 2
            write_fastq_sequence(options.output_file, *seq1)
            write_fastq_sequence(options.output_file, *seq2)
            wcount += 1
            report_counts(count, wcount, wcount)
            continue

        mate1[key1] = seq1
        mate2[key2] = seq2

        if key1 in mate2:
            write_fastq_sequence(options.output_file, *mate1[key1])
            write_fastq_sequence(options.output_file, *mate2[key1])
            del mate1[key1]
            del mate2[key1]
            wcount += 1
        if key2 in mate1:
            write_fastq_sequence(options.output_file, *mate1[key2])
            write_fastq_sequence(options.output_file, *mate2[key2])
            del mate1[key2]
            del mate2[key2]
            wcount += 1

        report_counts(count, wcount, count)

    report_counts(count, wcount, None)


def main():
    "Main function"
    options = set_parser().parse_args()

    # configs log and set log level
    mgkit.logger.config_log(options.verbose)
    options.func(options)

if __name__ == '__main__':
    main()



"""
Add more information to GFF annotations: gene mappings, coverage, taxonomy,
etc..

Uniprot Command
***************

If the *gene_id* of an annotation is a Uniprot ID, the script queries Uniprot
for the requested information. At the moment the information that can be added
is the taxon_id, taxon_name, lineage and mapping to EC, KO, eggNOG IDs.

It's also possible to add mappings to other databases using the *-m* option
with the correct identifier for the mapping, which can be found at `this page
<http://www.uniprot.org/faq/28>`_; for example if it's we want to add the
mappings of uniprot IDs to *BioCyc*, in the *abbreviation* column of the
mappings we find that it's identifier is *REACTOME_ID*, so we pass
*-m REACTOME* to the script (leaving *_ID* out). Mapped IDs are separated by
commas.

The taxonomy IDs are not overwritten if they are found in the annotations, the
*-f* is provided to force the overwriting of those values.

See also :ref:`gff-specs` for more informations about the GFF specifications
used.

.. note::

    As the script needs to query Uniprot a lot, it is recommended to split
    the GFF in several files, so an error in the connection doesn't waste time.

    However, a cache is kept to reduce the number of connections

Taxonomy Command
****************

To refine the taxonomic assignments of predicted genes annotations, the
annotation sequences may be searched against a database like the NCBI *nt*.

This commands takes as input a GFF file, one or more blast output files and a
file with all mappings from GIDs to taxonomy IDs. More information on how to
get the file can be read in the documentation of the function
:func:`mgkit.io.blast.parse_gi_taxa_table`.

The fasta sequences used with BLAST must have as name the uid of the
annotations they refer to, and one way to obtain these sequences is to use the
function :func:`mgkit.io.gff.extract_nuc_seqs` and save them to a fasta file.
Another options is to use the `sequence` command of the `get-gff-info` script
(:ref:`get-gff-info`).

The command accept a minimum bitscore to accept an hit and the taxon ID is
selected by default using top hit method, but LCA can be used, using the *-l*
switch.

Top Hit
+++++++

The best hit is selected from all those found for a sequence which has the
maximum bitscore and identity, with the bitscore having the highest priority.

LCA Taxon
+++++++++

Activated with the *-l* switch, it selects the last common ancestor of all
taxon IDs that are from the cellular organism root in the taxonomy and are
within a 10 bits (by default, can be customised with *-a*) from the hit with
the highest bitscore. If a taxon ID is not found in the taxonomy, it is
excluded. One of the requirements of this option is a file that contains the
full taxonomy from Uniprot/NCBI. The file can be obtained with the following
command::

    $ download_data -x -p -m your@email

The command will output a `taxonomy.pickle` file that can be passed to the `-x`
option :ref:`download-data`.

Coverage Command
****************

Adds coverage information from BAM alignment files to a GFF file, using the
function :func:`mgkit.align.add_coverage_info`, the user needs to supply for
each sample a BAM file, using the `-a` option, whose parameter is in the form
`sample,samplealg.bam`. More samples can be supplied adding more `-a`
arguments.

.. hint::

    As an example, to add coverage for `sample1`, `sample2` the command line
    is::

        add-gff-info coverage -a sample1,sample1.bam -a sample2,sample2.bam \\
        inputgff outputgff

A total coverage for the annotation is also calculated and stored in the
`cov` attribute, while each sample coverage is stored into `sample_cov` as per
:ref:`gff-specs`.

Adding Coverage from samtools depth
***********************************

The *cov_samtools* allows the use of the output of *samtools* **depth**
command. The *-aa* options must be used to pass information about all base
pairs and sequences coverage in the BAM/SAM file. The command accept only one
sample and the relative file/stream from *samtools*, meaning that multiple
samples coverage information must be added one at a time. One solution is to
pipe multiple commands to obtain result wanted. For example::

    $ add-gff-info cov_samtools -s SAMPLE1 -d sample1-coverage input.gff | \
        add-gff-info cov_samtools -s SAMPLE2 -d sample2-coverage - output.gff

This command will add the coverage information for SAMPLE1 and SAMPLE2 from
the respective files.

Uniprot Offline Mappings
************************

Similar to the *uniprot* command, it uses the `idmapping <ftp://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/idmapping/idmapping.dat.gz>`_
file provided by Uniprot, which speeds up the process of adding mappings and
taxonomy IDs from Uniprot gene IDs. It's not possible tough to add *EC*
mappings with this command, as those are not included in the file.

Kegg Information
****************

The *kegg* command allows to add information to each annotation. Right now the
information that can be added is restricted to the pathway(s) (reference KO) a
KO is part of and both the KO and pathway(s) descriptions. This information is
stored in keys starting with **ko_**.

Expected Aminoacidic Changes
****************************

Some scripts, like :ref:`snp-parser`, require information about the expected
number of synonymous and non-synonymous changes of an annotation. This can be
done using :meth:`mgkit.io.gff.Annotation.add_exp_syn_count` by the user of the
command `exp_syn` of this script. The attributes added to each annotation are
explained in the :ref:`gff-specs`

Adding Information from eggNOG
******************************

The *eggnog* command allows to add information from the *annotations* file
available for profiles in eggNOG.

Adding Count Data
*****************

Count data on a per-sample basis can be added with the *counts* command. The
accepted inputs are from HTSeq-count and featureCounts. The ouput produced by
featureCounts, is the one from using its **-f** option must be used.

This script accept by default a tab separated file, with a uid in the first
column and the other columns are the counts for each sample, in the same order
as they are passed to the **-s** option. To use the featureCounts file format,
this script **-e** option must be used.

The sample names must be provided in the same order as the columns in the input
files. If the counts are FPKMS the *-f* option can be used.

Adding Taxonomy from a Table
****************************

There are cases where it may needed or preferred to add the taxonomy from a
*gene_id* already provided in the GFF file. For such cases the *addtaxa*
command can be used. It works in a similar way to the *taxonomy* command, only
it expect three different type of inputs:

    * *GI-Taxa* table from NCBI (e.g. gi_taxid_nucl.dmp, )
    * tab separated table
    * dictionary

The first two are tab separated files, where on each line, the first column is
the *gene_id* that is found in the first column, while the second if the
*taxon_id*.

The last option is a serialised Python *dict*/hash table, whose keys are the
*gene_id* and the value is that gene corresponding *taxon_id*. The serialised
formats accepted are msgpack, json and pickle. The *msgpack* module must be
importable. The option to use json and msgpack allow to integrate this script
with other languages without resorting to a text file.

While the default is to look for the *gene_id* attribute in the GFF annotation,
another attribute can be specified, using the **-gene-attr** option.

.. note::

    the dictionary content is loaded after the table files and its keys and
    corresponding values takes precedence over the text files.

.. warning::

    from September 2016 NCBI will retire the GI. In that case the same
    kind of table can be built from the *nucl_gb.accession2taxid.gz* file
    The format is different, but some information can be found in
    :func:`mgkit.io.blast.parse_accession_taxa_table`


Adding information from Pfam
****************************

Adds the Pfam description for the annotation, by downloading the list from
Pfam.

The options allow to specify in which attribute the ID/ACCESSION is stored
(defaults to *gene_id*) and which one between ID/ACCESSION is the value of that
attribute (defaults to *ID*). if no description is found for the family, a
warning message is logged.

Changes
*******

.. versionchanged:: 0.3.0
    added *cov_samtools* command, *--split* option to *exp_syn*, *-c* option to
    *addtaxa*

.. versionchanged:: 0.2.6
    added *skip-no-taxon* option to *addtaxa*

.. versionchanged:: 0.2.5
    if a dictionary is supplied to *addtaxa*, the GFF is not preloaded

.. versionchanged:: 0.2.3
    added *pfam* command, renamed *gitaxa* to *addtaxa* and made it general

.. versionchanged:: 0.2.2
    added *eggnog*, *gitaxa* and *counts* command

.. versionchanged:: 0.2.1

* added *-d* to *uniprot* command
* added cache to *uniprot* command
* added *kegg* command (cached)

.. versionchanged:: 0.1.16
    added *exp_syn* command

.. versionchanged:: 0.1.15
    *taxonomy* command *-b* option changed

.. versionchanged:: 0.1.13

* added *--force-taxon-id* option to the *uniprot* command
* added *coverage* command
* added *taxonomy* command
* added *unipfile* command

.. versionadded:: 0.1.12
"""
from __future__ import division
import sys
import argparse
import logging
import itertools
import functools
import pysam
import json
import pickle
import mgkit
from . import utils
from .. import align
from .. import logger
from .. import taxon
from .. import kegg
from .. import DependencyError
from ..io import gff, blast, fasta, compressed_handle, open_file
from ..io import uniprot as uniprot_io
from ..net import uniprot as uniprot_net
from ..net import pfam
from ..utils.dictionary import cache_dict_file


LOG = logging.getLogger(__name__)


def set_common_options(parser):
    parser.add_argument(
        'input_file',
        nargs='?',
        type=argparse.FileType('r'),
        default='-',
        help='Input GFF file, defaults to stdin'
    )
    parser.add_argument(
        'output_file',
        nargs='?',
        type=argparse.FileType('w'),
        default=sys.stdout,
        help='Output GFF file, defaults to stdout'
    )


def set_kegg_parser(parser):
    parser.add_argument(
        '-c',
        '--email',
        action='store',
        type=str,
        help='Contact email',
        default=None
    )
    group = parser.add_argument_group('Requires Internet connection')
    group.add_argument(
        '-d',
        '--description',
        action='store_true',
        default=False,
        help='Add Kegg description'
    )
    group.add_argument(
        '-p',
        '--pathways',
        action='store_true',
        default=False,
        help='Add pathways ID involved'
    )
    group.add_argument(
        '-m',
        '--kegg-id',
        action='store',
        default='gene_id',
        help="""
        In which attribute the Kegg ID is stored (defaults to *gene_id*)"""
    )
    parser.set_defaults(func=kegg_command)


def kegg_command(options):

    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    kegg_client = kegg.KeggClientRest()

    LOG.info('Retrieving KO names')
    ko_names = kegg_client.get_names('ko')

    if options.pathways:
        LOG.info('Retrieving Pathways names')
        # Changes the names of the keys to *ko* instead of *map*
        path_names = dict(
            (path_id.replace('map', 'ko'), name)
            for path_id, name in kegg_client.get_names('path').iteritems()
        )

    ko_cache = {}

    for annotation in gff.parse_gff(options.input_file, gff_type=gff.from_gff):
        try:
            ko_id = annotation.attr[options.kegg_id]
        except KeyError:
            continue
        # if more than one KO is defined
        if ',' in ko_id:
            LOG.warning(
                'More than one KO assigned, skipping annotation: %s',
                annotation.uid
            )
            annotation.to_file(options.output_file)
            continue
        try:
            ko_info = ko_cache[ko_id]
        except KeyError:
            ko_info = {}
            if options.pathways:
                ko_info['pathway'] = kegg_client.link_ids('path', ko_id)
                # If left empty, no pathway will be saved
                if ko_info['pathway']:
                    ko_info['pathway'] = list(
                        path_id
                        for path_id in ko_info['pathway'][ko_id]
                        if path_id.startswith('ko')
                    )
            ko_cache[ko_id] = ko_info
        if options.description:
            annotation.attr['ko_description'] = ko_names.get(ko_id, ko_id)
        if options.pathways and ko_info['pathway']:
            annotation.attr['ko_pathway'] = ','.join(ko_info['pathway'])
            annotation.attr['ko_pathway_names'] = ','.join(
                path_names[path_id]
                for path_id in ko_info['pathway']
            )

        annotation.to_file(options.output_file)


def set_uniprot_parser(parser):
    parser.add_argument(
        '-c',
        '--email',
        action='store',
        type=str,
        help='Contact email',
        default=None
    )
    parser.add_argument(
        '--buffer',
        action='store',
        type=int,
        help='Number of annotations to keep in memory',
        default=50
    )
    group = parser.add_argument_group('Requires Internet connection')
    group.add_argument(
        '-f',
        '--force-taxon-id',
        action='store_true',
        default=False,
        help='Overwrite taxon_id if already present'
    )
    group.add_argument(
        '-t',
        '--taxon-id',
        action='store_true',
        default=False,
        help="""
             Add taxonomic ids to annotations, if taxon_id is found, it won't
             be Overwritten.
             """
    )
    group.add_argument(
        '-l',
        '--lineage',
        action='store_true',
        default=False,
        help='Add taxonomic lineage to annotations'
    )
    group.add_argument(
        '-e',
        '--eggnog',
        action='store_true',
        default=False,
        help='Add eggNOG mappings to annotations'
    )
    group.add_argument(
        '-ec',
        action='store_true',
        default=False,
        help='Add EC mappings to annotations'
    )
    group.add_argument(
        '-ko',
        action='store_true',
        default=False,
        help='Add KO mappings to annotations'
    )
    group.add_argument(
        '-d',
        '--protein-names',
        action='store_true',
        default=False,
        help='Add Uniprot description'
    )
    group.add_argument(
        '-m',
        '--mapping',
        action='append',
        type=str,
        help='Add any DB mappings to annotations'
    )

    parser.set_defaults(func=uniprot_command)


def add_uniprot_info(annotations, options, info_cache):
    columns = []

    if options.taxon_id:
        columns.append('organism')
        columns.append('organism-id')
    if options.eggnog:
        columns.append('database(EGGNOG)')
    if options.ko:
        columns.append('database(KO)')
    if options.ec:
        columns.append('ec')
    if options.lineage:
        columns.append('lineage(ALL)')
    if options.mapping is not None:
        for db in options.mapping:
            columns.append('database({0})'.format(db))
    if options.protein_names:
        columns.append('protein names')

    if not columns:
        return

    LOG.info("Retrieving gene information from Uniprot")

    gene_ids = set(
        x.gene_id
        for x in annotations
        if x.gene_id not in info_cache
    )

    # It may be that all gene_ids are already cached. In which case there's no
    # point querying Uniprot (it will probably give an error)
    if gene_ids:
        data = uniprot_net.get_gene_info(
            gene_ids,
            columns=columns,
            contact=options.email
        )

        info_cache.update(data)

    for annotation in annotations:
        try:
            gene_info = info_cache[annotation.gene_id]
        except KeyError:
            # no data was found
            continue

        for column, values in gene_info.iteritems():
            # nothing found
            if not values:
                continue

            if column == 'organism-id':
                if (annotation.taxon_id and options.force_taxon_id) or \
                   (annotation.taxon_id is None):
                    annotation.attr['taxon_id'] = int(values)
                    annotation.attr['taxon_db'] = 'UNIPROT'
                    # test with a try/expect maybe
                    if 'organism' in columns:
                        annotation.attr['taxon_name'] = gene_info['organism']
            elif column.startswith('lineage'):
                    annotation.attr['lineage'] = gene_info['lineage(ALL)']
            elif column.startswith('database'):
                annotation.attr[
                    'map_{0}'.format(
                        column[:-1].split('(')[1]
                    )
                ] = ','.join(values)
            elif column == 'ec':
                if isinstance(values, list):
                    annotation.attr['EC'] = ','.join(
                        x.split('-')[0]
                        for x in values
                    )
                else:
                    annotation.attr['EC'] = values
            elif column == 'protein names':
                annotation.attr['uniprot_description'] = values


def uniprot_command(options):

    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    if options.buffer < 1:
        options.buffer = 1

    ann_buffer = []

    info_cache = {}

    for annotation in gff.parse_gff(options.input_file, gff_type=gff.from_gff):

        ann_buffer.append(annotation)

        if len(ann_buffer) == options.buffer:

            add_uniprot_info(ann_buffer, options, info_cache)

            for annotation in ann_buffer:
                annotation.to_file(options.output_file)

            ann_buffer = []
    else:
        add_uniprot_info(ann_buffer, options, info_cache)

        for annotation in ann_buffer:
            annotation.to_file(options.output_file)


def set_blast_taxonomy_parser(parser):
    parser.add_argument(
        '-t',
        '--gi-taxa-table',
        action='store',
        default=None,
        required=True,
        help="GIDs taxonomy table (e.g. gi_taxid_nucl.dmp.gz)"
    )
    parser.add_argument(
        '-b',
        '--blast-output',
        action='append',
        default=None,
        required=True,
        help="BLAST output file(s)"
    )
    parser.add_argument(
        '-s',
        '--bitscore',
        action='store',
        default=40,
        type=float,
        help="Minimum bitscore allowed"
    )
    parser.add_argument(
        '-d',
        '--taxon-db',
        action='store',
        default='NCBI-NT',
        help="NCBI database used"
    )
    group = parser.add_argument_group('LCA options')
    group.add_argument(
        '-l',
        '--lca',
        action='store_true',
        default=False,
        help="Use last common ancestor to solve ambiguities"
    )
    group.add_argument(
        '-x',
        '--taxonomy',
        action='store',
        type=argparse.FileType('r'),
        default=None,
        help="Taxonomy file"
    )
    group.add_argument(
        '-a',
        '--max-diff',
        action='store',
        default=10,
        type=float,
        help="Bitscore difference from the max hit"
    )
    parser.set_defaults(func=taxonomy_command)


def get_gids(uid_gid_map):

    gids = set()

    for hits in uid_gid_map.itervalues():
        gids.update(x[0] for x in hits)

    return gids


def choose_by_lca(hits, taxonomy, gid_taxon_map, score=10):
    # the minimum score required to be part of the taxon IDs selected is to
    # be at most 10 bits (by default) from the maximum bitscore found in the
    # hits
    min_score = max(hits, key=lambda x: x[-1])[-1] - score

    # filter over only cellular organisms
    cellular_organism = 131567

    taxon_ids = set()

    for gid, identity, bitscore in hits:
        # No taxon_id found for the gid
        taxon_id = gid_taxon_map.get(gid, None)

        # skip already included taxon_ids
        if taxon_id in taxon_ids:
            continue

        # if the requirements are met
        if (taxon_id is not None) and (bitscore >= min_score):
            try:
                # test to make sure it is a cellular organism
                if taxon.is_ancestor(taxonomy, taxon_id, cellular_organism):
                    taxon_ids.add(taxon_id)
                # not part of the loop
                else:
                    LOG.debug("%d not part of cellular organisms", taxon_id)
            # The taxon id is not in the taxonomy used. It's skipped
            except KeyError:
                LOG.warning("%d is not part of the taxonomy", taxon_id)
                continue

    # no taxon_id passes the filters
    if len(taxon_ids) == 0:
        return None

    # log message fired up only if the package is set on DEBUG
    if mgkit.DEBUG and (len(taxon_ids) > 1):
        LOG.debug(
            "Num hits %d: %s",
            len(taxon_ids),
            ','.join(
                taxonomy[taxon_id].s_name
                for taxon_id in taxon_ids
                if taxon_id in taxonomy
            )
        )

    func = functools.partial(
        taxon.last_common_ancestor,
        taxonomy
    )

    return reduce(func, taxon_ids)


def choose_by_score(hits, gid_taxon_map):
    return gid_taxon_map.get(max(hits, key=lambda x: (x[-1], x[-2]))[0], None)


def taxonomy_command(options):

    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    if options.lca:
        LOG.info(
            "Using LCA to resolve multiple hits %.2f bits from the top hit",
            options.max_diff
        )
        if options.taxonomy is None:
            utils.exit_script('A taxonomy file is required', 1)
        options.taxonomy = taxon.UniprotTaxonomy(options.taxonomy)

    uid_gid_map = dict(
        itertools.chain(
            *(
                blast.parse_fragment_blast(x, bitscore=options.bitscore)
                for x in options.blast_output
            )
        )
    )

    gids = get_gids(uid_gid_map)

    gid_taxon_map = dict(
        (gid, taxon_id)
        for gid, taxon_id in blast.parse_gi_taxa_table(
            options.gi_taxa_table, gids=gids
        )
    )

    count = 0
    tot_count = 0

    for annotation in gff.parse_gff(options.input_file):
        tot_count += 1

        hits = uid_gid_map.get(annotation.uid, None)

        if hits is not None:
            if options.lca:
                taxon_id = choose_by_lca(
                    hits,
                    options.taxonomy,
                    gid_taxon_map,
                    options.max_diff
                )
                if taxon_id is not None:
                    LOG.debug(
                        "Selected %s (%s) by LCA",
                        options.taxonomy[taxon_id].s_name,
                        options.taxonomy[taxon_id].rank
                    )
            else:
                taxon_id = choose_by_score(
                    hits,
                    gid_taxon_map
                )
        else:
            taxon_id = None

        if taxon_id is not None:
            count += 1
            annotation.taxon_id = taxon_id
            annotation.taxon_db = options.taxon_db

        annotation.to_file(options.output_file)

    LOG.info(
        "Added taxonomy information to %.2f%% annotations (%d/%d)",
        count / tot_count * 100,
        count,
        tot_count
    )


def split_sample_alg(argument):
    "Split sample/alignment option"
    try:
        sample, bam_file_name = argument.split(',', 1)
    except ValueError:
        raise argparse.ArgumentTypeError(
            "Can't get get both sample and bam file from '%s'" % argument
        )

    return (sample, bam_file_name)


def set_coverage_parser(parser):
    parser.add_argument(
        '-a',
        '--sample-alignment',
        action='append',
        required=True,
        type=split_sample_alg,
        help='sample name and correspondent alignment file separated by comma'
    )
    parser.set_defaults(func=coverage_command)


def coverage_command(options):
    samples = []
    bam_files = []

    for sample, bam_file_name in options.sample_alignment:
        samples.append(sample)
        bam_files.append(pysam.Samfile(bam_file_name, 'rb'))

    if len(samples) != len(set(samples)):
        LOG.critical("There are duplicate sample names")
        return 1

    annotations = list(gff.parse_gff(options.input_file))

    align.add_coverage_info(annotations, bam_files, samples)

    gff.write_gff(annotations, options.output_file)


def set_exp_syn_parser(parser):
    """
    .. versionadded:: 0.1.16
    """
    parser.add_argument(
        '-r',
        '--reference',
        required=True,
        type=argparse.FileType('r'),
        help='reference sequence in fasta format'
    )
    parser.add_argument(
        '-s',
        '--split',
        action='store_true',
        help='''Split the sequence header of the reference at the first
        space, to emulate BLAST behaviour''',
        default=False
    )
    parser.set_defaults(func=exp_syn_command)


def exp_syn_command(options):

    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    """
    .. versionadded:: 0.1.16
    """

    seqs = dict(
        (
            seq_id.split(' ')[0] if options.split else seq_id,
            seq
        )
        for seq_id, seq in fasta.load_fasta(options.reference)
    )

    for annotation in gff.parse_gff(options.input_file):
        annotation.add_exp_syn_count(seqs[annotation.seq_id])

        annotation.to_file(options.output_file)


def uniprot_offline_command(options):

    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    LOG.info("Mappings selected: %s", ', '.join(options.mapping))

    annotations = []
    gene_ids = set()

    for annotation in gff.parse_gff(options.input_file):
        annotations.append(annotation)
        gene_ids.add(annotation.gene_id)

    iterator = uniprot_io.uniprot_mappings_to_dict(
        options.mapping_file,
        gene_ids=set(gene_ids),
        mappings=set(options.mapping)
    )

    file_mappings = dict(iterator)

    count = 0

    for annotation in annotations:
        try:
            mappings = file_mappings[annotation.gene_id]
            for mapping_id, mapping_values in mappings.iteritems():
                if mapping_id == uniprot_io.MAPPINGS['taxonomy']:
                    if (annotation.taxon_id and options.force_taxon_id) or \
                       (annotation.taxon_id is None):
                        annotation.taxon_id = mapping_values[0]
                        annotation.taxon_db = 'UNIPROT'
                else:
                    annotation.set_mapping(mapping_id, mapping_values)
            count += 1
        except KeyError:
            pass

        annotation.to_file(options.output_file)

    LOG.info(
        "Number of annotation changed: %d/%d (%.2f%%)",
        count,
        len(annotations),
        count / len(annotations) * 100
    )


def set_uniprot_offline_parser(parser):
    parser.add_argument(
        '-i',
        '--mapping-file',
        action='store',
        default='idmapping.dat.gz',
        required=True,
        help="Uniprot mapping file"
    )
    parser.add_argument(
        '-f',
        '--force-taxon-id',
        action='store_true',
        default=False,
        help="Overwrite taxon_id if already present"
    )
    parser.add_argument(
        '-m',
        '--mapping',
        action='append',
        default=None,
        required=True,
        choices=uniprot_io.MAPPINGS.values(),
        help="Mappings to add"
    )
    parser.set_defaults(func=uniprot_offline_command)


def read_lines_from_files(file_handles):
    file_handles = [
        compressed_handle(file_handle)
        if file_handle.name.endswith('gz')
        else file_handle

        for file_handle in file_handles
    ]
    for line in itertools.chain(*(x for x in file_handles)):
        # the __ is for HTSeq-count files, # for featureCounts
        if line.startswith('#') or line.startswith('__'):
            continue
        yield line.strip()


def eggnog_command(options):

    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    base_info = {}
    LOG.info(
        "Reading eggNOG annotations from files: %s",
        ', '.join(x.name for x in options.annotations_file)
    )
    for line in read_lines_from_files(options.annotations_file):
        level, gene_id, description, source = line.split('\t')
        base_info[gene_id] = dict(
            level=level,
            description=description,
            source=source
        )
    for annotation in gff.parse_gff(options.input_file):
        try:
            ann_info = base_info[annotation.gene_id]
            for key, value in ann_info.iteritems():
                annotation.set_attr(
                    'eggnog_{}'.format(key),
                    value
                )
        except KeyError:
            LOG.warning('No annotation for gene_id %s', annotation.gene_id)
        annotation.to_file(options.output_file)


def set_eggnog_parser(parser):
    parser.add_argument(
        '-a',
        '--annotations-file',
        action='append',
        required=True,
        type=argparse.FileType('r'),
        help="Annotations file"
    )
    parser.set_defaults(func=eggnog_command)


def load_counts(count_files, samples, featureCounts):
    LOG.info("Sample: %s", ', '.join(samples))

    counts = {}

    index = 6 if featureCounts else 1

    for line in read_lines_from_files(count_files):
        # featureCounts puts the header on the second line
        # the first is skipped by read_lines_from_files
        # as it starts with a #
        if featureCounts and line.startswith('Geneid'):
            continue
        line = line.split('\t')
        uid = line[0]
        if len(line[index:]) != len(samples):
            utils.exit_script(
                "The number of samples is not the same as the columns in " +
                "the count file",
                2
            )
        counts[uid] = {}
        for sample, count in zip(samples, line[index:]):
            count = float(count)
            # to save some memory
            if count == 0:
                continue
            counts[uid][sample] = count

    LOG.info("Loaded counts for %s annotations", len(counts))

    return counts


def counts_command(options):

    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    counts = load_counts(
        options.count_files,
        options.samples,
        options.featureCounts
    )

    key = "fpkms_{}" if options.fpkms else "counts_{}"

    for annotation in gff.parse_gff(options.input_file):
        try:
            ann_counts = counts[annotation.uid]
        except KeyError:
            LOG.warning("No counts found for annotation %s", annotation.uid)
            annotation.to_file(options.output_file)
            continue

        for sample in options.samples:
            annotation.attr[key.format(sample)] = ann_counts.get(sample, 0)

        annotation.to_file(options.output_file)


def set_counts_parser(parser):
    parser.add_argument(
        '-s',
        '--samples',
        action='store',
        required=True,
        type=lambda x: [y for y in x.split(',') if y],
        help="Comma separated sample names, in the same order as the count " +
             " file"
    )
    parser.add_argument(
        '-c',
        '--count-files',
        action='append',
        required=True,
        type=argparse.FileType('r'),
        help="Count file(s)"
    )
    parser.add_argument(
        '-f',
        '--fpkms',
        action='store_true',
        default=False,
        help="If the counts are FPKMS"
    )
    parser.add_argument(
        '-e',
        '--featureCounts',
        action='store_true',
        default=False,
        help="If the counts files are from featureCounts (using the -f option)"
    )
    parser.set_defaults(func=counts_command)


def addtaxa_command(options):

    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    if (options.gene_taxon_table is not None) and (not options.cache_table):
        annotations = []
        gene_ids = set()

        # the annotations are preloaded if a table is supplied, such as
        # the one from NCBI, so only the gene_ids necessary are taken from that
        # table to save memory

        for annotation in gff.parse_gff(options.input_file):
            annotations.append(annotation)
            gene_ids.add(annotation.attr[options.gene_attr])
        gene_ids = dict(
            blast.parse_accession_taxa_table(
                options.gene_taxon_table,
                acc_ids=gene_ids,
                key=0,
                value=1,
                no_zero=True
            )
        )
    elif (options.gene_taxon_table is not None) and options.cache_table:
        annotations = gff.parse_gff(options.input_file)
        gene_ids = cache_dict_file(
            blast.parse_accession_taxa_table(
                options.gene_taxon_table,
                key=0,
                value=1,
                no_zero=True
            )
        )
    else:
        # in case a dictionary is supplied, it's expected to fit in memory,
        # meaning that the GFF doesn't have to be preloaded
        annotations = gff.parse_gff(options.input_file)
        gene_ids = {}

    if options.dictionary is not None:
        dict_file = open_file(options.dictionary, 'r')
        if '.json' in options.dictionary:
            gene_ids.update(json.load(dict_file))
        elif '.msgpack':
            try:
                import msgpack
            except ImportError:
                raise DependencyError('msgpack')
            gene_ids.update(msgpack.load(dict_file))
        else:
            gene_ids.update(pickle.load(dict_file))

        # Ensures all taxon_ids are *int*
        gene_ids = dict(
            (key, int(value))
            for key, value in gene_ids.iteritems()
        )

    if options.taxonomy is not None:
        taxonomy = taxon.UniprotTaxonomy(options.taxonomy)

    for annotation in annotations:
        try:
            taxon_id = gene_ids[annotation.attr[options.gene_attr]]
            annotation.taxon_id = taxon_id
            annotation.set_attr('taxon_db', options.taxon_db)
        except KeyError:
            LOG.error(
                "No Taxon ID for GENE - %s",
                annotation.attr[options.gene_attr]
            )
            taxon_id = None
            if options.skip_no_taxon:
                continue

        if (options.taxonomy is not None) and (taxon_id is not None):
            try:
                annotation.attr['taxon_name'] = taxonomy[taxon_id].s_name
                annotation.attr['lineage'] = ','.join(
                    taxon_name
                    for taxon_name in taxon.get_lineage(taxonomy, taxon_id, names=True, only_ranked=True)
                    if taxon_name
                )
            except KeyError:
                LOG.warning("Taxon ID %d not found in the Taxonomy", taxon_id)

        annotation.to_file(options.output_file)


def set_addtaxa_parser(parser):
    parser.add_argument(
        '-t',
        '--gene-taxon-table',
        action='store',
        default=None,
        help="""GIDs taxonomy table (e.g. gi_taxid_nucl.dmp.gz) or a similar
                file where GENE/TAXON are tab separated and one per line"""
    )
    parser.add_argument(
        '-a',
        '--gene-attr',
        action='store',
        default='gene_id',
        help="""
        In which attribute the GENEID in the table is stored (defaults to
        *gene_id*)"""
    )
    parser.add_argument(
        '-x',
        '--taxonomy',
        action='store',
        default=None,
        help="""
        Taxonomy file - If given, both *taxon_name* and *lineage* attributes
        will be set
        """
    )
    parser.add_argument(
        '-d',
        '--dictionary',
        action='store',
        default=None,
        help="""
        A serialised dictionary, where the key is the GENEID and the value is
        TAXONID. It can be in json or msgpack format (can be a compressed file)
        *Note*: the dictionary values takes precedence over the table files
        """
    )
    parser.add_argument(
        '-e',
        '--skip-no-taxon',
        action='store_true',
        default=False,
        help="""If used, annotations with no taxon_id won't be outputted"""
    )
    parser.add_argument(
        '-db',
        '--taxon-db',
        action='store',
        default='NONE',
        help="DB used to add the taxonomic information"
    )
    parser.add_argument(
        '-c',
        '--cache-table',
        action='store_true',
        default=False,
        help="""If used, annotations are not preloaded, but the taxa table is
        cached, instead.
        """
    )

    parser.set_defaults(func=addtaxa_command)


def pfam_command(options):

    LOG.info(
        'Writing to file (%s)',
        getattr(options.output_file, 'name', repr(options.output_file))
    )

    LOG.info("Downloading Pfam Information")

    pfam_families = pfam.get_pfam_families(
        'acc' if options.use_accession else 'id'
    )

    for annotation in gff.parse_gff(options.input_file):
        pfam_id = annotation.attr[options.id_attr]
        try:
            annotation.attr['pfam_description'] = pfam_families[pfam_id][1]
        except KeyError:
            LOG.warning("No description found for family %s", pfam_id)
        annotation.to_file(options.output_file)


def set_pfam_parser(parser):
    parser.add_argument(
        '-i',
        '--id-attr',
        action='store',
        default='gene_id',
        help="""
        In which attribute the Pfam ID/ACCESSION is stored (defaults to
        *gene_id*)"""
    )
    parser.add_argument(
        '-a',
        '--use-accession',
        action='store_true',
        default=False,
        help="""If used, the attribute value is the Pfam ACCESSION
        (e.g. PF06894), not ID (e.g. Phage_TAC_2)"""
    )
    parser.set_defaults(func=pfam_command)


def set_samtools_depth_parser(parser):
    parser.add_argument(
        '-s',
        '--sample',
        action='store',
        type=str,
        help='sample name'
    )
    parser.add_argument(
        '-d',
        '--depth',
        action='store',
        required=True,
        type=argparse.FileType('r'),
        help='`samtools depth -aa` file'
    )
    parser.add_argument(
        '-n',
        '--num-seqs',
        action='store',
        default=10**4,
        type=int,
        help='Number of sequences to update the log'
    )
    parser.set_defaults(func=samtools_depth_command)


def samtools_depth_command(options):
    depth = align.SamtoolsDepth(options.depth, options.num_seqs)
    if options.sample is None:
        cov_attr = 'cov'
    else:
        cov_attr = '{}_cov'.format(options.sample)

    for annotation in gff.parse_gff(options.input_file):
        cov = depth.region_coverage(
            annotation.seq_id,
            annotation.start,
            annotation.end
        )
        annotation.set_attr(cov_attr, cov)
        annotation.to_file(options.output_file)


def set_parser():
    """
    Sets command line arguments parser
    """
    parser = argparse.ArgumentParser(
        description='Adds informations to a GFF file',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    subparsers = parser.add_subparsers()

    parser_u = subparsers.add_parser(
        'uniprot',
        help='Adds information from GFF whose gene_id is from Uniprot'
    )

    set_uniprot_parser(parser_u)
    set_common_options(parser_u)
    utils.add_basic_options(parser_u, manual=__doc__)

    parser_t = subparsers.add_parser(
        'taxonomy',
        help='''Adds taxonomic information from annotation sequences blasted
                against a NCBI db'''
    )

    set_blast_taxonomy_parser(parser_t)
    set_common_options(parser_t)
    utils.add_basic_options(parser_t, manual=__doc__)

    parser_c = subparsers.add_parser(
        'coverage',
        help='Adds coverage information from BAM Alignment files'
    )

    set_coverage_parser(parser_c)
    set_common_options(parser_c)
    utils.add_basic_options(parser_c, manual=__doc__)

    parser_e = subparsers.add_parser(
        'exp_syn',
        help='Adds expected synonymous and non-synonymous changes information'
    )

    set_exp_syn_parser(parser_e)
    set_common_options(parser_e)
    utils.add_basic_options(parser_e, manual=__doc__)

    parser_f = subparsers.add_parser(
        'unipfile',
        help='Adds mappings and taxonomy from Uniprot mapping file'
    )

    set_uniprot_offline_parser(parser_f)
    set_common_options(parser_f)
    utils.add_basic_options(parser_f, manual=__doc__)

    parser_k = subparsers.add_parser(
        'kegg',
        help='Adds information and mapping from Kegg'
    )

    set_kegg_parser(parser_k)
    set_common_options(parser_k)
    utils.add_basic_options(parser_k, manual=__doc__)

    parser_eggnog = subparsers.add_parser(
        'eggnog',
        help='Adds information from eggNOG'
    )

    set_eggnog_parser(parser_eggnog)
    set_common_options(parser_eggnog)
    utils.add_basic_options(parser_eggnog, manual=__doc__)

    parser_counts = subparsers.add_parser(
        'counts',
        help='Adds counts data to the GFF'
    )

    set_counts_parser(parser_counts)
    set_common_options(parser_counts)
    utils.add_basic_options(parser_counts, manual=__doc__)

    parser_addtaxa = subparsers.add_parser(
        'addtaxa',
        help='''Adds taxonomy information from a GI-Taxa, gene_id/taxon_id
                table or a dictionary serialised as a pickle/msgpack/json file
                '''
    )

    set_addtaxa_parser(parser_addtaxa)
    set_common_options(parser_addtaxa)
    utils.add_basic_options(parser_addtaxa, manual=__doc__)

    parser_pfam = subparsers.add_parser(
        'pfam',
        help='Adds information from Pfam'
    )

    set_pfam_parser(parser_pfam)
    set_common_options(parser_pfam)
    utils.add_basic_options(parser_pfam, manual=__doc__)

    parser_samtools_depth = subparsers.add_parser(
        'cov_samtools',
        help='Adds information from samtools_depth'
    )

    set_samtools_depth_parser(parser_samtools_depth)
    set_common_options(parser_samtools_depth)
    utils.add_basic_options(parser_samtools_depth, manual=__doc__)

    # top parser
    utils.add_basic_options(parser, manual=__doc__)

    return parser


def main():
    "Main function"

    options = set_parser().parse_args()

    logger.config_log(options.verbose)
    options.func(options)



"""
.. versionadded:: 0.1.12

Graph module
"""

import itertools
from . import DependencyError

try:
    import networkx as nx
except ImportError:
    raise DependencyError('networkx')


def build_graph(id_links, name, edge_type='', weight=0.5):
    """
    .. versionadded:: 0.1.12

    Builds a networkx graph from a dictionary of nodes, as outputted by
    :meth:`mgkit.kegg.KeggClientRest.get_pathway_links`. The graph is
    undirected, and all edges weight are the same.

    Arguments:
        id_links (dict): dictionary with the links
        name (str): name of the graph
        edge_type (str): an optional name for the `edge_type` attribute
            set for each edge
        weight (float): the weight assigned to each edge in the graph

    Returns:
        graph: an instance of :class:`networkx.Graph`
    """
    g = nx.Graph()
    g.name = name

    for id1, id2s in id_links.iteritems():
        g.add_node(id1, id=id1)
        for id2 in id2s:
            g.add_node(id2, id=id2)
            g.add_edge(id1, id2, edge_type=edge_type, weight=weight)

    return g


def build_weighted_graph(id_links, name, weights, edge_type=''):
    """
    .. versionadded:: 0.1.14

    Builds a networkx graph from a dictionary of nodes, as outputted by
    :meth:`mgkit.kegg.KeggClientRest.get_pathway_links`. The graph is
    undirected, and all edges weight are the same.

    Arguments:
        id_links (dict): dictionary with the links
        name (str): name of the graph
        edge_type (str): an optional name for the `edge_type` attribute
            set for each edge
        weight (float): the weight assigned to each edge in the graph

    Returns:
        graph: an instance of :class:`networkx.Graph`
    """
    g = nx.Graph()
    g.name = name

    for id1, id2s in id_links:
        g.add_node(id1, id=id1)
        for id2 in id2s:
            g.add_node(id2, id=id2)
            try:
                weight = weights[(id1, id2)]
            except KeyError:
                weight = weights.get(id1, 0.5)
            g.add_edge(id1, id2, edge_type=edge_type, weight=float(weight))

    return g


def rename_graph_nodes(graph, name_func=None, exclude_ids=None):
    new_graph = nx.Graph()

    if exclude_ids is None:
        exclude_ids = set()

    for node, data in graph.nodes_iter(data=True):
        new_graph.add_node(
            node if node in exclude_ids else name_func(node),
            **data
        )

    for node1, node2, data in graph.edges_iter(data=True):
        new_graph.add_edge(
            node1 if node1 in exclude_ids else name_func(node1),
            node2 if node2 in exclude_ids else name_func(node2),
            **data
        )

    return new_graph


def copy_nodes(g, graph1, name=None, id_attr=None, **kwd):
    """
    .. versionadded:: 0.1.12

    Used by :func:`link_nodes` to copy nodes
    """

    if name is None:
        name = graph1.name

    if id_attr is None:
        id_attr = 'id'

    for node, data in graph1.nodes_iter(data=True):
        data = data.copy()
        data.update(kwd)
        g.add_node(
            "{0}_{1}".format(name, data[id_attr]),
            **data
        )


def copy_edges(g, graph1, name=None, **kwd):
    """
    .. versionadded:: 0.1.12

    Used by :func:`link_nodes` to copy edges
    """

    if name is None:
        name = graph1.name

    for node1, node2, data in graph1.edges_iter(data=True):
        data = data.copy()
        data.update(kwd)
        g.add_edge(
            "{0}_{1}".format(name, node1),
            "{0}_{1}".format(name, node2),
            **data
        )


def link_nodes(g, graph1, graph2, id_filter, link_type, weight):
    """
    .. versionadded:: 0.1.12

    Used by :func:`link_graph` to link nodes with the same *id*
    """
    nodes1 = set(
        data['id']
        for node, data in graph1.nodes_iter(data=True)
        if id_filter(data['id'])
    )
    nodes2 = set(
        data['id']
        for node, data in graph2.nodes_iter(data=True)
        if id_filter(data['id'])
    )

    for node_id in (nodes1 & nodes2):
        g.node["{0}_{1}".format(graph1.name, node_id)]['alpha'] = 0.5
        g.node["{0}_{1}".format(graph2.name, node_id)]['alpha'] = 0.5
        g.add_edge(
            "{0}_{1}".format(graph1.name, node_id),
            "{0}_{1}".format(graph2.name, node_id),
            edge_type=link_type,
            weight=weight
        )

EDGE_LINKS = [
    (lambda x: x.startswith('C'), 'CP_LINK', 0.0),
    (lambda x: x.startswith('K'), 'KO_LINK', 0.0)
]
"Sample edge_links for :func:`link_graph`"


def link_graph(graphs, edge_links):
    """
    .. versionadded:: 0.1.12

    Link nodes of a set of graphs using the specifics in edge_links.
    The resulting graph nodes are renamed, and the nodes that are shared
    between the graphs linked.

    Arguments:
        graphs: iterable of graphs
        edge_links: iterable with function, edge_type and weight for the
            links between graphs

    Returns:
        graph: an instance of :class:`networkx.Graph`
    """
    g = nx.Graph()

    for graph in graphs:
        copy_nodes(g, graph)
        copy_edges(g, graph)

    for graph1, graph2 in itertools.combinations(graphs, r=2):
        for id_filter, link_type, weight in edge_links:
            link_nodes(g, graph1, graph2, id_filter, link_type, weight)

    return g


def filter_graph(graph, id_list, filter_func=lambda x: x.startswith('K')):
    """
    .. versionadded:: 0.1.12

    Filter a graph based on the `id_list` provided and the filter function
    used to test the id attribute of each node.

    A node is removed if `filter_func` returns True on a node and its id
    attribute is not in `id_list`

    Arguments:
        graph: the graph to filter
        id_list (iterable): the list of nodes that are to remain in the
            graph
        filter_func (func): function which accept a single parameter and
            return a boolean

    Returns:
        graph: an instance of :class:`networkx.Graph`
    """
    graph = graph.copy()
    nodes = [
        node
        for node, data in graph.nodes_iter(data=True)
        if filter_func(data['id']) and (data['id'] not in id_list)
    ]

    graph.remove_nodes_from(nodes)

    graph.remove_nodes_from(nx.isolates(graph))

    return graph


def annotate_graph_nodes(graph, attr, id_map, default=None):
    """
    .. versionadded:: 0.1.12

    Add/Changes nodes attribute `attr` using a dictionary of ids->values.

    .. note::

        If the id is not found in `id_map`:

        * default is None: no value added for that node
        * default is not None: the node attribute will be set to `default`

    Arguments:
        graph: the graph to annotate
        attr (str): the attribute to annotate
        id_map (dict): the dictionary with the values for each node
        default: the value used in case an `id` is not found in `id_map`
    """
    for node, data in graph.nodes_iter(data=True):
        try:
            data[attr] = id_map[data['id']]
        except KeyError:
            if default is not None:
                data[attr] = default



"""
Translate nucleotidic sequence in amino acidic sequences
"""

import argparse
import logging
import itertools
from .. import logger
from ..io.fasta import load_fasta, write_fasta_sequence
from ..utils import trans_tables
from ..utils.sequence import translate_sequence
from . import utils

LOG = logging.getLogger(__name__)

try:
    from joblib import Parallel, delayed
except ImportError:
    Parallel = None


def set_parser():
    "argument parser configuration"
    parser = argparse.ArgumentParser(
        description='Translate sequences from nucleotidic to amino acidic',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        'input_file',
        action='store',
        help='input file with aa sequences',
        type=argparse.FileType('r')
    )
    parser.add_argument(
        'output_file',
        action='store',
        help='output file with aa sequences',
        type=argparse.FileType('w')
    )
    parser.add_argument(
        '-t',
        '--trans-table',
        default='universal',
        action='store',
        choices=[
            table_name.lower() for table_name in dir(trans_tables)
            if not table_name.startswith('_')
        ],
        help='translation table'
    )
    parser.add_argument(
        '-p',
        '--processors',
        action='store',
        default=1,
        type=int,
        help='Number of processors to use'
    )
    parser.add_argument(
        '-n',
        '--buffer-size',
        action='store',
        default=50000,
        type=int,
        help='Number of sequences to read/write at a time'
    )
    utils.add_basic_options(parser, manual=__doc__)

    return parser


def load_trans_table(table_name):
    "Loads translation table "
    return getattr(trans_tables, table_name.upper())


def translate_seq(name, seq, trans_table):
    "Tranlate sequence into the 6 frames"
    seqs = []
    header = "{0}-{1}{2}"
    for start in range(3):
        seqs.append(
            (header.format(name, 'f', start),
             translate_sequence(seq, start, trans_table, False))
        )
        seqs.append(
            (header.format(name, 'r', start),
             translate_sequence(seq, start, trans_table, True))
        )
    return seqs


def translate_buff(jobs, buff_seqs, trans_table):
    "Process a set amount of sequences - multiprocess"

    aa_seqs = jobs(
        delayed(translate_seq)(name, seq, trans_table)
        for name, seq in buff_seqs
    )

    # LOG.info("Finished translating buffer")

    return aa_seqs


def main():
    "Main function"
    options = set_parser().parse_args()

    # configs log and set log level
    logger.config_log(options.verbose)

    trans_table = load_trans_table(options.trans_table)

    nuc_seqs = list(load_fasta(options.input_file))

    if Parallel is not None:
        LOG.info("Using %d processor(s)", options.processors)
        jobs = Parallel(n_jobs=options.processors, verbose=0)

    for index in xrange(0, len(nuc_seqs), options.buffer_size):
        LOG.info(
            "Translating file position %d-%d",
            index,
            index+options.buffer_size
        )
        if Parallel is None:
            aa_seqs = [
                translate_seq(name, seq, trans_table)
                for name, seq in nuc_seqs[index:index+options.buffer_size]
            ]
        else:
            aa_seqs = translate_buff(
                jobs,
                nuc_seqs[index:index+options.buffer_size],
                trans_table
            )
        for name, seq in itertools.chain(*aa_seqs):
            write_fasta_sequence(options.output_file, name, seq)

    options.output_file.close()

    return 0

if __name__ == '__main__':
    main()




"""
Module containing classes and functions to deal with Gene Ontology data
"""

import urllib2
import logging
from .. import kegg
from ..utils import dictionary as dict_utils

LOG = logging.getLogger(__name__)

# needed to complete the documentation without raising an exception
try:
    from goatools.obo_parser import OBOReader
except ImportError:
    LOG.critical("Cannot import OBOReader")


class Kegg2GOMapper(kegg.KeggMapperBase):
    """
    Class holding mappings KOs->GO
    """
    query_string = 'database:(type:ko {0})'
    columns_string = 'go-id'
    _id_names = {}

    def map_kos_go(self, kos, contact, skip=False):
        """
        Map a list of KOs to GO ids
        """
        msg1 = "(%05d/%05d) Already downloaded mapping for KO: %s - %s"
        msg2 = "(%05d/%05d) Getting GO mappings for KO: %s - %s"
        msg3 = "(%05d/%05d) Poblem downloading KO %s (%s)"

        errors = 0
        for idx, ko_id in enumerate(kos):
            if skip and (ko_id in self._ko_map or ko_id in self._not_found):
                LOG.info(msg1, idx + 1, len(kos), ko_id, kos[ko_id])
                continue
            LOG.info(msg2, idx + 1, len(kos), ko_id, kos[ko_id])
            try:
                mappings = self.ko_to_mapping(ko_id, self.query_string,
                                              self.columns_string, contact)
            except urllib2.HTTPError as url_error:
                LOG.warning(msg3, idx + 1, len(kos), ko_id, url_error.args)
                errors += 1
                continue
            if mappings:
                self._ko_map[ko_id] = mappings
            else:
                self._not_found.append(ko_id)
        return errors

    def load_goslim_mapping(self, f_handle):
        if isinstance(f_handle, str):
            f_handle = open(f_handle, 'r')

        goslim_map = {}

        for line in f_handle:
            line = line.split('//')[0]
            goterm, goslim = line.split('=>')
            goterm = goterm.strip()
            if not goterm.startswith('GO'):
                continue
            goslim = goslim.strip().split()
            goslim_map[goterm] = goslim

        self._ko_map = dict_utils.combine_dict(self._ko_map, goslim_map)

    def load_go_names(self, fname):
        name_map = self._id_names
        for term in OBOReader(fname):
            name_map[term.id] = term.name

    def write_association_file(self, f_handle):
        if isinstance(f_handle, str):
            f_handle = open(f_handle, 'w')

        ko_map = self.get_ko_map()

        for ko_id, goterms in ko_map.iteritems():
            f_handle.write("{0}\t{1}\n".format(ko_id, ';'.join(goterms)))


def download_data(contact, kegg_data='kegg.pickle', go_data='go_data.pickle'):
    """
    Function to download Gene Ontology data
    """

    g_obj = Kegg2GOMapper()
    try:
        g_obj.load_data(go_data)
    except IOError:
        LOG.warning("GO data not found, download restarting")

    LOG.info("Loading Kegg data from file %s", kegg_data)
    kegg_obj = kegg.KeggData(kegg_data)
    ko_names = kegg_obj.get_ko_names()

    try:
        errors = g_obj.map_kos_go(ko_names, contact, skip=True)
    finally:
        LOG.info("Saving downloaded data to %s", go_data)
        g_obj.save_data(go_data)

    LOG.info(
        "Found %d/%d mappings. Number of errors: %d",
        len(g_obj), len(ko_names), errors
    )



"""
This module gives access to Uniprot taxonomy data. It also defines classes
to filter, order and group data by taxa
"""

import logging
import cPickle
import itertools
import collections
from .io import open_file
from . import DependencyError
# from .utils.common import deprecated


LOG = logging.getLogger(__name__)

ALGAE = {
    'haptophyceae': 2830,
    'chlorophyta': 3041,
    'stramenopiles': 33634,
    'cryptophyta': 3027,
    'rhodophyta': 2763,
}

# superkingdoms
BACTERIA = 2
ARCHAEA = 2157
VIRUS = 10239

CELLULAR_ORGANISMS = 131567
# kingdoms
FUNGI = 4751
VIRIDIPLANTAE = 33090
METAZOA = 33208
EUKARYOTA = 2759

PROTISTS = {
    #'apicomplexa', alveolata
    #'ciliophora', alveolata
    'alveolata': 33630,
    'amoebozoa': 554915,
    'apusozoa': 554296,
    'breviatea': 1401294,
    'centroheliozoa': 193537,
    'choanoflagellida': 28009,
    'diplomonadida': 5738,
    'euglenozoa': 33682,
    'formicata': 207245,
    'heterolobosea': 5752,
    'jakobida': 556282,
    'malawimonadidae': 136087,
    'oxymonadida': 66288,
    'parabasalia': 5719,
    'rhizaria': 543769,
}

PLANTS = {
    'streptophyta': 35493,
}

TAXON_ROOTS = (
    'archaea',  # 2157
    'bacteria',  # 2
    'fungi',  # 4751
    'metazoa',
    'environmental samples',
    'viruses',
    'viroids'
    'eukaryota',
    'other sequences',
    'unidentified'
) + tuple(PROTISTS) + tuple(PLANTS) + tuple(ALGAE)
"Root taxa used in analysis and filtering"

TAXON_RANKS = (
    'superkingdom',
    'kingdom',
    'phylum',
    'class',
    'subclass',
    'order',
    'family',
    'genus',
    'species'
)
"Taxonomy ranks included in the pickled data"

UniprotTaxonTuple = collections.namedtuple(
    'UniprotTaxonTuple',
    ('taxon_id', 's_name', 'c_name', 'rank', 'lineage', 'parent_id')
)
"""
A representation of a Uniprot Taxon
"""


def parse_uniprot_taxon(line, light=True):
    """
    .. versionchanged:: 0.1.13
        now accepts empty scientific names, for root taxa

    .. versionchanged:: 0.2.1
        added *light* parameter

    Parses a Uniprot taxonomy file (tab delimited) line and returns a
    UniprotTaxonTuple instance. If *light* is True, lineage is not stored to
    decrease the memory usage. This is now the default.
    """
    line = line.rstrip().split('\t')
    taxon_id = int(line[0])
    try:
        s_name = line[2].lower()
    except IndexError:
        s_name = ''
    try:
        c_name = line[3].lower() if line[3] else ''
    except IndexError:
        c_name = ''
    try:
        rank = line[7].lower()
    except IndexError:
        rank = None
    try:
        if light:
            raise IndexError
        lineage = tuple(x.lower() for x in line[8].split('; '))
    except IndexError:
        lineage = (None,)
    try:
        parent_id = int(line[9])
    except (ValueError, IndexError):
        parent_id = None
    return UniprotTaxonTuple(
        taxon_id,
        s_name,
        c_name,
        rank,
        lineage,
        parent_id
    )


def parse_ncbi_taxonomy_merged_file(file_handle):
    """
    .. versionadded:: 0.2.3

    Parses the *merged.dmp* file where the merged taxon_id are stored. Available
    at ftp://ftp.ncbi.nlm.nih.gov/pub/taxonomy/

    Arguments:
        file_handle (str, file): file name or handle to the file

    Returns:
        dict: dictionary with merged_id -> taxon_id
    """
    file_handle = open_file(file_handle)

    LOG.info(
        "Reading NCBI taxonomy merged file %s",
        getattr(file_handle, 'name', repr(file_handle))
    )

    merged_taxa = {}

    for line in file_handle:

        merged_id, taxon_id = [col for col in line.strip().split('\t') if col != '|']
        merged_taxa[int(merged_id)] = int(taxon_id)

    return merged_taxa


def parse_ncbi_taxonomy_names_file(file_handle, name_classes=('scientific name', 'common name')):
    """
    .. versionadded:: 0.2.3

    Parses the *names.dmp* file where the names associated to a taxon_id are
    stored. Available at ftp://ftp.ncbi.nlm.nih.gov/pub/taxonomy/

    Arguments:
        file_handle (str, file): file name or handle to the file
        name_classes (tuple): name classes to save, only the scientific and
            common name are stored

    Returns:
        dict: dictionary with merged_id -> taxon_id
    """

    file_handle = open_file(file_handle)

    LOG.info(
        "Reading NCBI taxonomy names file %s",
        getattr(file_handle, 'name', repr(file_handle))
    )

    taxa_names = {}

    for line in file_handle:
        taxon_id, taxon_name, uniq_name, name_class = [
            col.strip()
            for col in line.strip().split('\t')
            if col != '|'
        ]
        taxon_id = int(taxon_id)

        if name_class not in name_classes:
            continue

        if taxon_id not in taxa_names:
            taxa_names[taxon_id] = {}

        taxa_names[taxon_id][name_class] = taxon_name

    return taxa_names


def parse_ncbi_taxonomy_nodes_file(file_handle, taxa_names=None):
    """
    .. versionadded:: 0.2.3

    Parses the *nodes.dmp* file where the nodes of the taxonomy are stored.
    Available at ftp://ftp.ncbi.nlm.nih.gov/pub/taxonomy/.

    Arguments:
        file_handle (str, file): file name or handle to the file
        taxa_names (dict): dictionary with the taxa names (returned from
            :func:`parse_ncbi_taxonomy_names_file`)

    Yields:
        UniprotTaxonTuple: UniprotTaxonTuple instance
    """

    file_handle = open_file(file_handle)

    LOG.info(
        "Reading NCBI taxonomy nodes file %s",
        getattr(file_handle, 'name', repr(file_handle))
    )

    for line in file_handle:
        line = [col for col in line.strip().split('\t') if col != '|']
        taxon_id = int(line[0])
        parent_id = int(line[1])
        if parent_id == 1:
            # NCBI uses 1 as the root, but the rest of the functions expect None
            parent_id = None
        rank = line[2].lower()

        s_name = ''
        c_name = ''

        if taxa_names is not None:
            try:
                names = taxa_names[taxon_id]
            except KeyError:
                names = {}
            s_name = names.get('scientific name', '')
            c_name = names.get('common name', '')

        yield UniprotTaxonTuple(
            taxon_id,
            s_name,
            c_name,
            rank,
            (None,),  # lineage is not found in the NCBI taxonomy dump
            parent_id
        )


class UniprotTaxonomy(object):
    """
    Class that contains the whole Uniprot taxonomy. Defines some methods to
    easy access of taxonomy.

    Defines:

    * methods to load taxonomy from a pickle file or a generic file handle
    * can be iterated over and returns a generator its UniprotTaxon instances
    * can be used as a dictionary, in which the key is a taxon_id and the value
      is its UniprotTaxon instance
    """
    def __init__(self, fname=None):
        """
        :param str fname: file name of the pickled data.
        """
        self._taxa = {}
        self._name_map = {}
        if fname:
            self.load_data(fname)

    def read_from_gtdb_taxonomy(self, file_handle, use_gtdb_name=True, sep='\t'):
        """
        .. versionadded:: 0.3.0

        Reads a GTDB taxonomy file (tab separated genome_id/taxonomy) and
        populate the taxonomy instance. The method also return a dictionary of
        genome_id -> taxon_id.

        Arguments:
            file_handle (file): file with the taxonomy
            use_gtdb_name (bool): if True, the names are kept as-is in the
                *s_name* attribute of :class:`UniprotTaxonTuple` and the
                "cleaned" version in *c_name* (e.g. f__Ammonifexaceae ->
                Ammonifexaceae). If False, the values are switched
            sep (str): separator between the columns of the file

        Returns:
            dict: dictionary of genome_id -> taxon_id, reflecting the created
            taxonomy

        .. note::

            the taxon_id are generated, so there's no guarantee they will be
            the same in a successive execution

        """
        if isinstance(file_handle, str):
            file_handle = open_file(file_handle)

        LOG.info(
            "Reading GTDB taxonomy from file",
            getattr(file_handle, 'name', repr(file_handle))
        )

        genome_ids = {}

        # ranks used in GTDB
        ranks = {
            'g': 'genus',
            'f': 'family',
            'p': 'phylum',
            'o': 'order',
            'd': 'domain',
            's': 'species',
            'c': 'class'
        }

        taxon_ids = {}
        count = 1
        for line in file_handle:
            # expecting the table to be the exported file from GTDB
            try:
                # print line.strip().split(sep)
                genome_id, line = line.strip().split(sep)
                # in case there's no gtdb taxonomy
            except ValueError:
                continue

            line = [
                taxon_name
                for taxon_name in line.split(';')
                if len(taxon_name) > 3
            ]

            if not line:
                continue

            # if no parent, use None as the NCBI taxonomy
            parent_id = None
            for taxon_name in line:

                # register a taxo_id for unknown taxon
                if taxon_name not in taxon_ids:
                    taxon_ids[taxon_name] = count
                    count += 1

                taxon_id = taxon_ids[taxon_name]
                try:
                    rank = ranks[taxon_name[0]]
                except KeyError:
                    raise KeyError

                # keep the full gtdb name in the common name
                # cut the scientific name to remove the rank information
                common_name = taxon_name
                taxon_name = taxon_name[3:]
                # but if th use_gtdb_name option is True, switch them
                if use_gtdb_name:
                    taxon_name, common_name = common_name, taxon_name

                self[taxon_id] = UniprotTaxonTuple(
                    taxon_id,
                    taxon_name,
                    common_name,
                    rank,
                    (None,),
                    parent_id
                )
                parent_id = taxon_id
            genome_ids[genome_id] = taxon_id

        return genome_ids

    def read_from_ncbi_dump(self, nodes_file, names_file=None, merged_file=None):
        """
            .. versionadded:: 0.2.3

            Uses the *nodes.dmp* and optionally *names.dmp*, *merged.dmp* files
            from ftp://ftp.ncbi.nlm.nih.gov/pub/taxonomy/ to populate the
            taxonomy.

            Arguments:
                nodes_file (str, file): file name or handle to the file
                names_file (str, file, None): file name or handle to the file,
                    if None, names won't be added to the taxa
                merged_file (str, file, None): file name or handle to the file,
                    if None, pointers to merged taxa won't be added
            """

        LOG.info("Reading NCBI taxonomy from file")

        if names_file is not None:
            taxa_names = parse_ncbi_taxonomy_names_file(names_file)
        else:
            taxa_names = None

        for taxon in parse_ncbi_taxonomy_nodes_file(nodes_file, taxa_names):
            self[taxon.taxon_id] = taxon

        if merged_file is not None:
            merged_taxa = parse_ncbi_taxonomy_merged_file(merged_file)

            for merged_id, taxon_id in merged_taxa.iteritems():
                self[merged_id] = self[taxon_id]

    def read_taxonomy(self, f_handle, light=True):
        """
        .. versionchanged:: 0.2.1
            added *light* parameter

        Reads taxonomy from a file handle.
        The file needs to be a tab separated format return by a query on
        Uniprot.  If *light* is True, lineage is not stored to decrease the
        memory usage. This is now the default.

        New taxa will be added, duplicated taxa will be skipped.

        :param handle f_handle: file handle of the taxonomy file.
        """
        try:
            LOG.info("Loading taxonomy from file %s", f_handle.name)
        except AttributeError:
            pass

        f_handle.readline()

        for line in f_handle:
            try:
                # taxon = UniprotTaxon(line=line)
                taxon = parse_uniprot_taxon(line, light=light)
            except Exception:
                # print line.strip().split('\t')
                LOG.debug("skipped line: %s", line)
                continue
            if taxon.taxon_id in self:
                continue
            self[taxon.taxon_id] = taxon

    def gen_name_map(self):
        """
        .. versionchanged:: 0.2.3
            names are stored in the mapping as lowercase

        Generate a name map, where to each scientific name in the taxonomy an
        id is associated.
        """
        LOG.debug('Generate name map')
        self._name_map = {}
        for taxon_obj in self:
            s_name = taxon_obj.s_name.lower()
            try:
                self._name_map[s_name].append(taxon_obj.taxon_id)
            except KeyError:
                self._name_map[s_name] = [taxon_obj.taxon_id]

    def load_data(self, file_handle):
        """
        .. versionchanged:: 0.2.3
            now can use read *msgpack* serialised files

        .. versionchanged:: 0.1.13
            now accepts file handles and compressed files (if file names)

        Loads serialised data from file name "file_handle" and accepts
        compressed files.

        if the *.msgpack* string is found in the file name, the *msgpack*
        package is used instead of pickle

        Arguments:
            file_handle (str, file): file name to which save the instance data

        Raises:
            DependencyError: if the file name contains *.msgpack* and the
            package is not installed
        """

        if isinstance(file_handle, str):
            file_handle = open_file(file_handle)

        LOG.info("Loading taxonomy from file %s", file_handle.name)

        if '.msgpack' in file_handle.name:
            try:
                import msgpack
            except ImportError:
                raise DependencyError('msgpack')
            merged = []
            for taxon_id, taxon in msgpack.Unpacker(file_handle, use_list=False):
                taxon = UniprotTaxonTuple(*taxon)
                # if it's a merged taxon_id keep it and don't add it
                if taxon_id != taxon.taxon_id:
                    merged.append((taxon_id, taxon.taxon_id))
                else:
                    self[taxon_id] = taxon
            # the list of merged ids is iterated over and pointers are added to
            # the correct ids
            for merged_id, taxon_id in merged:
                self[merged_id] = self[taxon_id]

        else:
            self._taxa = cPickle.load(file_handle)

    def save_data(self, file_handle):
        """
        .. versionchanged:: 0.2.3
            now can use *msgpack* to serialise

        Saves taxonomy data to a file handle or file name, can write compressed
        data if the file ends with ".gz", ".bz2"

        if the *.msgpack* string is found in the file name, the *msgpack*
        package is used instead of pickle

        Arguments:
            file_handle (str, file): file name to which save the instance data

        Raises:
            DependencyError: if the file name contains *.msgpack* and the
            package is not installed
        """

        if isinstance(file_handle, str):
            file_handle = open_file(file_handle, 'w')

        LOG.info("Saving taxonomy to file %s", file_handle.name)

        if '.msgpack' in file_handle.name:
            try:
                import msgpack
            except ImportError:
                raise DependencyError('msgpack')
            for taxon_id, taxon in self._taxa.iteritems():
                file_handle.write(
                    msgpack.packb((taxon_id, taxon))
                )
        else:
            cPickle.dump(self._taxa, file_handle, -1)

    def find_by_name(self, s_name):
        """
        .. versionchanged:: 0.2.3
            the search is now case insensitive

        Returns the taxon IDs associated with the scientific name provided

        :param str s_name: the scientific name

        :return list: a reference to the list of IDs that have for `s_name`
        """
        if not self._name_map:
            self.gen_name_map()
        return self._name_map[s_name.lower()]

    def is_ancestor(self, leaf_id, anc_ids):
        """
        .. versionchanged:: 0.1.13
            now uses :func:`is_ancestor` and changed behavior

        Checks if a taxon is the leaf of another one, or a list of taxa.

        :param int leaf_id: leaf taxon id
        :param int anc_ids: ancestor taxon id(s)

        :return bool: True if the ancestor taxon is in the leaf taxon lineage

        """

        if isinstance(anc_ids, int):
            anc_ids = [anc_ids]

        for anc_id in anc_ids:
            if is_ancestor(self, leaf_id, anc_id):
                return True
        return False

    def get_ranked_taxon(self, taxon_id, rank=None, ranks=TAXON_RANKS, roots=False):
        """
        .. versionchanged:: 0.1.13
            added *roots* argument

        Traverse the branch of which the *taxon* argument is the leaf backward,
        to get the specific rank to which the *taxon* belongs to.

        .. warning::

            the *roots* options is kept for backward compatibility and should be
            be set to *False*

        :param taxon_id: id of the taxon or instance of :class:`UniprotTaxon`
        :param str rank: string that specify the rank, if None, the first valid
            rank will be searched. (i.e. the first with a value different from '')
        :param ranks: tuple of all taxonomy ranks, default to the default module
            value
        :param bool roots: if True, uses :data:`TAXON_ROOTS` to solve the root
            taxa
        :return: instance of :class:`UniprotTaxon` for the rank found.
        """

        if isinstance(taxon_id, int):
            ranked = self[taxon_id]
        else:
            raise ValueError("Not a valid taxon_id: {0}".format(taxon_id))

        while True:
            if (rank is None) and (ranked.rank != ''):
                break
            elif (ranked.rank == rank):
                break
            elif (ranked.rank in ranks) and \
                    (ranks.index(ranked.rank) < ranks.index(rank)):
                break
            elif ranked.parent_id is None:
                break
            # kept only for backward compatibility.
            # needs a check to other scripts
            elif roots and (ranked.s_name in TAXON_ROOTS):
                break

            ranked = self[ranked.parent_id]

        return ranked

    def get_name_map(self):
        "Returns a taxon_id->s_name dictionary"

        return dict(
            (taxon.taxon_id, taxon.s_name)
            for taxon in self
        )

    def get_taxon_root(self, taxon_id, roots=TAXON_ROOTS):
        """
        .. deprecated:: 0.2.6

        Given a :class:`UniprotTaxon` instance and the associated taxonomy, returns
        the correct root taxon the supplied taxon belongs to.

        :param int taxon_id: taxon id or instance of :class:`UniprotTaxon`
        :param iterable roots: list of root taxa scientific names

        :return: root :class:`UniprotTaxon` instance
        """

        if isinstance(taxon_id, int):
            taxon = self[taxon_id]
        else:
            raise ValueError("Not a valid taxon_id: {0}".format(taxon_id))

        while taxon.s_name not in roots:
            taxon = self[taxon.parent_id]

        return taxon

    def __getitem__(self, taxon_id):
        """
        Defines dictionary behavior. Key is a taxon_id, the returned value is a
        UniprotTaxon instance
        """
        return self._taxa[taxon_id]

    def __setitem__(self, taxon_id, taxon_obj):
        self._taxa[taxon_id] = taxon_obj

    def __delitem__(self, taxon_id):
        del self._taxa[taxon_id]

    def __iter__(self):
        """
        Defines iterable behavior. Returns a generator for UniprotTaxon instances
        """
        for taxon in self._taxa.itervalues():
            yield taxon

    def __len__(self):
        """
        Returns the number of taxa contained
        """
        return len(self._taxa)

    def __contains__(self, taxon):
        """
        Returns True if the taxon is in the taxonomy

        Accepts an int (check for taxon_id) or an instance of UniprotTaxon
        """
        if isinstance(taxon, int):
            return taxon in self._taxa
        return False

    def __repr__(self):
        """
        .. versionadded:: 0.2.5
        """
        return "{} - {} taxa".format(self.__class__, len(self))


def group_by_root(taxa, roots=TAXON_ROOTS, only_names=False, replace_space='#'):
    """
    .. deprecated:: 0.2.6

    Returns a dictionary containing as keys the root taxa and as values a list
    of taxa belonging to that group (checks lineage attribute of
    :class:`UniprotTaxon` instances)

    :param iterable taxa: iterable of :class:`UniprotTaxon` instances
    :param roots: root taxa to be used to construct the dictionary, defaults to
        the 5 defined in :data:`TAXON_ROOTS`
    :param bool only_names: boolean that specify what data type to return for
        values, if True returns strings (taxa names), if False
        :class:`UniprotTaxon` instances
    :param str replace_space: if only_names is True replaces spaces with the
        choosen character

    :return: a dictionary root->[taxa_ids]
    """
    groups = dict(
        (root, []) for root in roots
    )
    if not only_names:
        taxa.gen_name_map()
        for root in roots:
            groups[root].append(taxa.find_by_name(root)[0])

    for taxon in taxa:
        for root in roots:
            if root in taxon.lineage:
                groups[root].append(
                    taxon.s_name.replace(
                        ' ', replace_space) if only_names else taxon.taxon_id
                )
                break

    return groups


def load_taxonomy_map(taxon_file):
    "Loads taxonomy from file and return a map root_taxon->taxa"
    taxonomy = UniprotTaxonomy(taxon_file)
    return group_by_root(taxonomy, only_names=True)


def get_ancestor_map(leaf_ids, anc_ids, taxonomy):
    """
    This function returns a dictionary where every leaf taxon is associated
    with the right ancestors in anc_ids

    ex. {clostridium: [bacteria, clostridia]}
    """
    anc_map = {}

    for leaf_id in leaf_ids:
        for anc_id in anc_ids:
            if taxonomy.is_ancestor(leaf_id, anc_id) or (leaf_id == anc_id):
                try:
                    anc_map[leaf_id].add(anc_id)
                except KeyError:
                    anc_map[leaf_id] = set()
                    anc_map[leaf_id].add(anc_id)

    return anc_map


def is_ancestor(taxonomy, taxon_id, anc_id):
    """
    .. versionchanged:: 0.1.16
        if a *taxon_id* raises a KeyError, False is returned

    Determine if the given taxon id (taxon_id) has anc_id as ancestor.

    :param :class:`UniprotTaxonomy` taxonomy: taxonomy used to test
    :param int taxon_id: leaf taxon to test
    :param int anc_id: ancestor taxon to test against

    :return bool: True if anc_id is an ancestor of taxon_id or their the same
    """
    if taxon_id == anc_id:
        return True

    while True:
        # if a taxon_id is not found, ancestry is invalid anyway, so it returns
        # False
        try:
            taxon_id = taxonomy[taxon_id].parent_id
        except KeyError:
            return False
        # print taxon_id, anc_id
        if anc_id == taxon_id:
            return True
        if taxon_id is None:
            return False


class NoLcaFound(Exception):
    """
    .. versionadded:: 0.1.13

    Raised if no lowest common ancestor can be found in the taxonomy
    """
    pass


def last_common_ancestor(taxonomy, taxon_id1, taxon_id2):
    """
    .. versionadded:: 0.1.13

    Finds the last common ancestor of two taxon IDs. An alias to this function
    is in the same module, called *lowest_common_ancestor* for compatibility.

    Arguments:
        taxonomy: :class:`UniprotTaxonomy` instance used to test
        taxon_id1 (int): first taxon ID
        taxon_id2 (int): second taxon ID

    Raturns:
        int: taxon ID of the lowest common ancestor

    Raises:
        NoLcaFound: if no common ancestor can be found
    """
    lca_id = taxon_id1

    while True:
        if is_ancestor(taxonomy, taxon_id2, lca_id):
            break
        else:
            lca_id = taxonomy[lca_id].parent_id
            if lca_id is None:
                raise NoLcaFound('No common ancestry')

    return lca_id

lowest_common_ancestor = last_common_ancestor


def distance_taxa_ancestor(taxonomy, taxon_id, anc_id):
    """
    .. versionadded:: 0.1.16

    Function to calculate the distance between a taxon and the given ancestor

    The distance is equal to the number of step in the taxonomy taken to arrive
    at the ancestor.

    Arguments:
        taxonomy: :class:`UniprotTaxonomy` instance
        taxon_id (int): taxonomic identifier
        anc_id (int): taxonomic identifier of the ancestor

    Raturns:
        int: distance between *taxon_id* and it ancestor *anc_id*

    """
    distance = 0

    while True:
        if anc_id == taxon_id:
            break
        taxon_id = taxonomy[taxon_id].parent_id
        distance += 1

    return distance


def distance_two_taxa(taxonomy, taxon_id1, taxon_id2):
    """
    .. versionadded:: 0.1.16

    Calculate the distance between two taxa. The distance is equal to the sum
    steps it takes to traverse the taxonomy until their last common ancestor.

    Arguments:
        taxonomy: :class:`UniprotTaxonomy` instance
        taxon_id1 (int): taxonomic identifier of first taxon
        taxon_id2 (int): taxonomic identifier of second taxon

    Raturns:
        int: distance between *taxon_id1* and *taxon_id2*
    """
    lca_id = last_common_ancestor(taxonomy, taxon_id1, taxon_id2)

    return sum(
        distance_taxa_ancestor(taxonomy, taxon_id, lca_id)
        for taxon_id in (taxon_id1, taxon_id2)
    )


def taxa_distance_matrix(taxonomy, taxon_ids):
    """
    .. versionadded:: 0.1.16

    Given a list of taxonomic identifiers, returns a distance matrix in a
    pairwise manner by using :func:`distance_two_taxa` on all possible
    two element combinations of *taxon_ids*.

    Arguments:
        taxonomy: :class:`UniprotTaxonomy` instance
        taxon_ids (iterable): list taxonomic identifiers

    Returns:
        pandas.DataFrame: matrix with the pairwise distances of all *taxon_ids*
    """
    try:
        import pandas
    except ImportError:
        raise DependencyError('pandas')

    matrix = pandas.DataFrame(index=taxon_ids, columns=taxon_ids).fillna(0)

    for taxon_id1, taxon_id2 in itertools.combinations(taxon_ids, 2):
        distance = distance_two_taxa(taxonomy, taxon_id1, taxon_id2)
        matrix.loc[taxon_id1, taxon_id2] = distance
        matrix.loc[taxon_id2, taxon_id1] = distance

    return matrix


def get_lineage(taxonomy, taxon_id, names=False, only_ranked=False,
                with_last=False):
    """
    .. versionadded:: 0.2.1

    .. versionchanged:: 0.2.5
        added *only_ranked*

    .. versionchanged:: 0.3.0
        added *with_last*

    Returns the lineage of a taxon_id, as a list of taxon_id or taxa names

    Arguments:
        taxonomy: a :class:`UniprotTaxonomy` instance
        taxon_id (int): taxon_id whose lineage to return
        names (bool): if True, the returned list contains the names of the taxa
            instead of the taxon_id
        only_ranked (bool): if True, only taxonomic levels whose rank is in
            data:`TAXON_RANKS` will be returned
        with_last (bool): if True, the passed taxon_id is included in the
            lineage

    Returns:
        list: lineage of the taxon_id, the elements are `int` if names is False,
        and `str` when *names* is True. If a taxon has no scientific name, the
        common name is used. If *only_ranked* is True, the returned list only
        contains ranked taxa (according to :data:`TAXON_RANKS`).
    """
    lineage = []

    if with_last:
        lineage.append(taxon_id)

    while True:
        taxon_id = taxonomy[taxon_id].parent_id
        if taxon_id is None:
            break
        if only_ranked and (taxonomy[taxon_id].rank not in TAXON_RANKS):
            continue
        lineage.append(taxon_id)

    if names:
        lineage = [
            taxonomy[x].s_name if taxonomy[x].s_name else taxonomy[x].c_name
            for x in lineage
        ]

    return list(reversed(lineage))


def last_common_ancestor_multiple(taxonomy, taxon_ids):
    """
    .. versionadded:: 0.2.5

    Applies :func:`last_common_ancestor` to an iterable that yields *taxon_id*
    while removing any *None* values. If the list is of one element, that
    *taxon_id* is returned.

    Arguments:
        taxonomy: instance of :class:`UniprotTaxonomy`
        taxon_ids (iterable): an iterable that yields taxon_id

    Returns:
        int: the *taxon_id* that is the last common ancestor of all taxon_ids
        passed

    Raises:
        NoLcaFound: when no common ancestry is found or the number of
        *taxon_ids* is 0
    """

    taxon_ids = [
        taxon_id
        for taxon_id in taxon_ids
        if taxon_id is not None
    ]

    if not taxon_ids:
        raise NoLcaFound('Empty *taxon_ids*')

    return reduce(
        lambda x1, x2: last_common_ancestor(taxonomy, x1, x2),
        set(taxon_ids)
    )



"""
Module containing classes and functions to deal with eggNOG data

.. todo::

    * unify download of data from web

"""
from __future__ import print_function

from .. import kegg
import urllib2
import pickle
import logging
import gzip
import cStringIO
import itertools
from ..io import open_file
from ..utils import dictionary

LOG = logging.getLogger(__name__)

EGGNOG_CAT_NAMES = (
    'INFORMATION STORAGE AND PROCESSING',
    'CELLULAR PROCESSES AND SIGNALING',
    'METABOLISM',
    'POORLY CHARACTERIZED'
)
"""
Functional categories (broader)
"""

EGGNOG_CAT_KEYS = (
    ('J', 'A', 'K', 'L', 'B'),
    ('D', 'Y', 'V', 'T', 'M', 'N', 'Z', 'W', 'U', 'O'),
    ('C', 'G', 'E', 'F', 'H', 'I', 'P', 'Q'),
    ('R', 'S')
)
"""
Used to build map of broader categories (:data:`EGGNOG_CAT_NAMES`) to more
specific ones
"""

EGGNOG_CAT_MAP = dict(
    (label, categories)
    for label, categories in zip(EGGNOG_CAT_NAMES, EGGNOG_CAT_KEYS)
)
"""
Functional categories (broader, :data:`EGGNOG_CAT_NAMES`) mappings to more
specific one (:data:`EGGNOG_CAT`).
"""

EGGNOG_CAT = {

    # INFORMATION STORAGE AND PROCESSING
    'J': "Translation, ribosomal structure and biogenesis",
    'A': "RNA processing and modification",
    'K': "Transcription",
    'L': "Replication, recombination and repair",
    'B': "Chromatin structure and dynamics",

    # CELLULAR PROCESSES AND SIGNALING
    'D': "Cell cycle control, cell division, chromosome partitioning",
    'Y': "Nuclear structure",
    'V': "Defense mechanisms",
    'T': "Signal transduction mechanisms",
    'M': "Cell wall/membrane/envelope biogenesis",
    'N': "Cell motility",
    'Z': "Cytoskeleton",
    'W': "Extracellular structures",
    'U': "Intracellular trafficking, secretion, and vesicular transport",
    'O': "Posttranslational modification, protein turnover, chaperones",

    # METABOLISM
    'C': "Energy production and conversion",
    'G': "Carbohydrate transport and metabolism",
    'E': "Amino acid transport and metabolism",
    'F': "Nucleotide transport and metabolism",
    'H': "Coenzyme transport and metabolism",
    'I': "Lipid transport and metabolism",
    'P': "Inorganic ion transport and metabolism",
    'Q': "Secondary metabolites biosynthesis, transport and catabolism",

    # POORLY CHARACTERIZED
    'R': "General function prediction only",
    'S': "Function unknown"
}
"""
Single letter functional categories
"""


def get_general_eggnog_cat(category):
    """
    .. versionadded:: 0.1.14

    Returns the functional category (:data:`EGGNOG_CAT_NAMES` keys)
    for the requested single letter functional category (:data:`EGGNOG_CAT`
    keys)
    """
    return set(
        gen_category
        for gen_category, categories in EGGNOG_CAT_MAP.iteritems()
        if category in categories
    )


class Kegg2NogMapper(kegg.KeggMapperBase):
    """
    Usage

    to add a ko_id to the mapper use:
    * map_ko_to_eggnog(ko_id) which calls:

        * get_uniprot_from_ko
        * get_eggnog_from_uniprot

    return None if there was no problem or a dictionary with the exception
    thrown

    to make mappings ko_id -> eggnog-categories:
    * get_cat_mapping_from_file (file are in the current directory)
    * gen_ko_to_cat

    .. todo::

        * add method to load kos from file
        * optimise querying (like Kegg2CazyMapper)

    """
    query_string = 'database:(type:ko {0}) AND database:(type:eggnog)'
    columns_string = 'database(eggnog)'
    _egg_to_cat = None
    _ko_to_cat = None

    def __init__(self, fname=None, gen_ko_to_cat=True):
        super(Kegg2NogMapper, self).__init__()
        self._egg_to_cat = {}
        self._ko_to_cat = {}
        self._id_names = EGGNOG_CAT
        if fname:
            self.load_data(fname)
            if gen_ko_to_cat:
                self.gen_ko_to_cat()

    def map_ko_to_eggnog(self, ko_id, contact):
        if (ko_id in self._ko_map) or (ko_id in self._not_found):
            return
        try:
            mappings = self.ko_to_mapping(
                ko_id, self.query_string,
                self.columns_string, contact
            )
        except urllib2.HTTPError:
            return

        if mappings:
            self._ko_map[ko_id] = mappings
        else:
            self._not_found.append(ko_id)

    def map_kos_eggnog(self, kos, contact):
        errors = 0

        for idx, ko_id in enumerate(kos):
            LOG.info(
                "(%05d/%05d) Getting eggNOG mappings for KO: %s - %s", idx + 1,
                len(kos), ko_id, kos[ko_id]
            )
            try:
                self.map_ko_to_eggnog(ko_id, contact)
            except urllib2.HTTPError:
                LOG.warning("Problem getting mappings %s", ko_id)
                errors += 1

        return errors

    def get_cat_mapping_from_file(self, files=('COG.funccat.txt', 'NOG.funccat.txt', 'KOG.funccat.txt')):
        """
        Duplicate functionality of load_func_cat

        Categories are single letter and their descriptioons are in
        :data:`EGGNOG_CAT`, while the broader categories :data:`EGGNOG_CAT_MAP`
        """
        for fname in files:
            f = open(fname, 'r')
            for line in f:
                egg_id, cats = line.rstrip().split()
                self._egg_to_cat[egg_id] = list(cats)

    def get_cat_mapping(self, f_handle):
        for line in f_handle:
            line = line.rstrip()
            if line == '':
                continue
            egg_id, cats = line.rstrip().split('\t', 1)
            self._egg_to_cat[egg_id] = list(cats)

    def gen_ko_to_cat(self):
        for ko_id, egg_ids in self._ko_map.iteritems():
            cats = set()
            for egg_id in egg_ids:
                try:
                    cats.update(self._egg_to_cat[egg_id])
                except KeyError:
                    # some egg_ids have no assigned category
                    pass
            if cats:
                self._ko_to_cat[ko_id] = cats

    def get_ko_cat(self, ko_id):
        return sorted(self._ko_to_cat[ko_id])

    def get_ko_cat_dict(self):
        return self._ko_to_cat.copy()

    def get_ko_map(self):
        return self.get_ko_cat_dict()

    def get_cat_ko_dict(self):
        cat_map = {}

        for ko_id in self._ko_to_cat:
            for mapped_id in self._ko_to_cat[ko_id]:
                try:
                    cat_map[mapped_id].add(ko_id)
                except KeyError:
                    cat_map[mapped_id] = set([ko_id])
        return cat_map

    def save_data(self, fname):
        LOG.info("Saving data to %s", fname)
        pickle.dump(
            (self._ko_map, self._egg_to_cat, self._not_found), open(fname, 'w')
        )

    def load_data(self, fname):
        LOG.info("Loading data from %s", fname)
        self._ko_map, self._egg_to_cat, self._not_found = pickle.load(
            open(fname, 'r')
        )


def download_data(contact, kegg_data='kegg.pickle',
                  eggnog_data='eggnog.pickle',
                  base_url='http://eggnog.embl.de/version_3.0/data/downloads/'):

    egg = Kegg2NogMapper()
    try:
        egg.load_data(eggnog_data)
    except:
        LOG.warning("Eggnog data not found, download restarting")

    kdata = kegg.KeggData(kegg_data)
    ko_names = kdata.get_ko_names()

    urls = ('COG.funccat.txt.gz', 'KOG.funccat.txt.gz', 'NOG.funccat.txt.gz')

    try:
        errors = egg.map_kos_eggnog(ko_names, contact)

        for url in urls:
            LOG.info("Loading categories mapping from %s%s", base_url, url)
            f_handle = gzip.GzipFile(
                fileobj=cStringIO.StringIO(urllib2.urlopen(base_url + url).read()),
                mode='r'
            )
            egg.get_cat_mapping(f_handle)
    finally:
        egg.save_data(eggnog_data)

    LOG.info("Number of errors: %d", errors)


def print_ko_to_cat(data_file='eggnog.pickle', descr=False):
    e = Kegg2NogMapper()
    e.load_data(data_file)
    e.gen_ko_to_cat()

    for ko_id in sorted(e):
        try:
            cats = e.get_ko_cat(ko_id)
        except KeyError:
            cats = []

        print(
            "{0}\t{1}".format(
                ko_id,
                '\t'.join(
                    EGGNOG_CAT[x] if descr else x
                    for x in cats
                )
            )
        )


class NOGInfo(object):
    """
    .. versionadded:: 0.1.14

    Mappings from Uniprot to eggNOG

    ..note::

        load_description is optional
    """
    _map_nog_func = None
    "eggNOG COG/NOG to functional category dictionary"
    _map_nog_gene = None
    "eggNOG COG/NOG to gene id dictionary"
    _map_nog_desc = None
    "eggNOG id to description dictionary"
    _map_gene_nog = None
    "eggNOG gene id to COG/NOG dictionary"

    def __init__(self):
        self._map_nog_gene = {}
        self._map_gene_nog = {}
        self._map_nog_desc = {}
        self._map_nog_func = {}

    def load_members(self, file_handle):
        """
        Loads data from *NOG.members.txt.gz*

        *file_handle* can either an open file or a path
        """
        if isinstance(file_handle, str):
            file_handle = open_file(file_handle, 'r')

        LOG.info("Reading Members from %s", file_handle.name)

        map_nog_gene = {}

        for line in file_handle:

            if line.startswith('#'):
                continue

            nog_id, gene_id, start, end = line.strip().split()

            try:
                map_nog_gene[nog_id].append(gene_id)
            except KeyError:
                map_nog_gene[nog_id] = [gene_id]

        self._map_nog_gene.update(map_nog_gene)
        self._map_gene_nog.update(dictionary.reverse_mapping(map_nog_gene))

    def load_description(self, file_handle):
        """
        Loads data from *NOG.description.txt.gz*

        *file_handle* can either an open file or a path
        """
        if isinstance(file_handle, str):
            file_handle = open_file(file_handle, 'r')

        LOG.info("Reading Descriptions from %s", file_handle.name)

        map_nog_desc = {}

        for line in file_handle:

            try:
                nog_id, nog_desc = line.strip().split('\t', 1)
            except ValueError:
                nog_id = line.strip()
                nog_desc = ''

            map_nog_desc[nog_id] = nog_desc

        self._map_nog_desc.update(map_nog_desc)

    def load_funccat(self, file_handle):
        """
        Loads data from *NOG.funccat.txt.gz*

        *file_handle* can either an open file or a path
        """
        if isinstance(file_handle, str):
            file_handle = open_file(file_handle, 'r')

        LOG.info("Reading Functional Categories from %s", file_handle.name)

        map_nog_func = {}

        for line in file_handle:
            nog_id, nog_func = line.strip().split()
            nog_func = set(nog_func)

            map_nog_func[nog_id] = nog_func

        self._map_nog_func.update(map_nog_func)

    def get_nog_funccat(self, nog_id):
        """
        Returns the functional category (one letter, :data:`EGGNOG_CAT` keys)
        for the requested eggNOG COG/NOG ID
        """
        try:
            return self._map_nog_func[nog_id].copy()
        except KeyError:
            return set()

    def get_nogs_funccat(self, nog_ids):
        """
        Returns the functional categories for a list of COG/NOG IDs. Uses
        :meth:`NOGInfo.get_nog_funccat`
        """

        iterator = (self.get_nog_funccat(nog_id) for nog_id in nog_ids)

        return set(itertools.chain(*iterator))

    def get_nog_gencat(self, nog_id):
        """
        Returns the functional category (:data:`EGGNOG_CAT_NAMES` keys)
        for the requested eggNOG COG/NOG ID
        """
        return set(
            itertools.chain(
                *(
                    get_general_eggnog_cat(funccat)
                    for funccat in self.get_nog_funccat(nog_id)
                )
            )
        )

    def get_gene_nog(self, gene_id):
        """
        Returns the COG/NOG ID of the requested eggNOG gene ID
        """
        try:
            nog_ids = self._map_gene_nog[gene_id]
        except KeyError:
            return None

        return nog_ids

    def get_gene_funccat(self, gene_id):
        """
        Returns the functional category (one letter, :data:`EGGNOG_CAT` keys)
        for the requested eggNOG gene ID
        """
        try:
            nog_ids = self._map_gene_nog[gene_id]
        except KeyError:
            return None

        return self.get_nogs_funccat(nog_ids)



"""
Workflow associated with assembly statistics and evaluation
"""

import logging
from ..io import gff
from .. import taxon
from ..utils.sequence import calc_n50
import numpy

LOG = logging.getLogger(__name__)


def rank_annotations_by_attr(annotations, attr='taxon'):
    """
    For all annotations in the list (usually all annotations for a contig),
    counts how many time a set attribute 'attr' appears. The resulting
    dictionary is then sorted by the number of counts and the one with the
    highest count is ranked by how much it represent the total number of
    counts.

    The rank is an integer number between 0 and 10.

    :param iterable annotations: list of :class:`gff.GFFKegg` instances
    :param str attr: the attribute for which the annotations are counted

    :return tuple: the attr with the most counts and its rank
    """
    taxon_counts = {}

    for annotation in annotations:
        taxon_name = getattr(annotation.attributes, attr)
        try:
            taxon_counts[taxon_name] += 1
        except KeyError:
            taxon_counts[taxon_name] = 1

    taxon_counts = sorted(
        (count, taxon_name) for taxon_name, count in taxon_counts.items()
    )
    total_counts = float(sum(x[0] for x in taxon_counts))

    rank = int((taxon_counts[-1][0] / total_counts) * 10)

    return taxon_counts[-1][1], rank


def assign_contigs_to_taxa(annotations, root_map=None,
                           black_list=None):
    """
    Groups annotations by contig (seq_id) and counts how many contigs a taxon,
    or its root if root_map is supplied, have been assigned to.

    The actual form of the dictionary like this:

    .. digraph:: results
        :alt: taxon->ranks->count

        "taxon 1" -> "rank 0" -> count;
        "taxon 1" -> "rank 1" -> count;
        "taxon 1" -> "rank 4" -> count;
        "taxon 1" -> "rank 7" -> count;
        "taxon 1" -> "rank 9" -> count;
        "taxon 1" -> "rank 10" -> count;

    .. note::

        the number of ranks for a taxon is not pretedermined, but depends on
        the values returned by :func:`rank_annotations_by_attr`.

    :param iterable annotations: list of :class:`gff.GFFKegg` instances
    :param dict root_map: dictionary taxon->root

    :return dict: dictionary
    """

    LOG.info("Assigning %d annotations", len(annotations))
    LOG.info(
        "Blacklisted taxa: %s",
        'none' if root_map is None else ', '.join(black_list)
    )
    if root_map is not None:
        LOG.info("Using only root taxa")

    contigs = gff.group_by_contig(annotations)

    assignments = {}
    for contig, ann_list in contigs.iteritems():
        taxon_name, rank = rank_annotations_by_attr(ann_list, attr='taxon')

        if black_list is not None:
            if taxon_name in black_list:
                continue

        if root_map is not None:
            taxon_name = taxon.get_taxon_root(taxon_name, root_map)

        if taxon_name not in assignments:
            assignments[taxon_name] = {}
        value = 1
        try:
            assignments[taxon_name][rank] += value
        except KeyError:
            assignments[taxon_name][rank] = value

    return assignments


def filter_contig_assignments(contig_assign, threshold=5, min_counts=1):
    """
    Filter contigs assignments using a threshold for the rank: all rank counts
    belonging to a taxon which are greater than or equal to threshold will be
    summed up.

    :param dict contig_assign: dictionary returned by
        :func:`assign_contigs_to_taxa`
    :param int threshold: the minimum rank for which the counts are summed up

    :return dict: dictionary in the form taxon_name->count
    """

    LOG.info(
        "Filtering contig assignments with a minimum rank of %d and a " +
        "minimum count of %d", threshold, min_counts
    )

    new_counts = {}
    for taxon, rank_dict in contig_assign.items():

        rank_sum = sum(
            count for rank, count in rank_dict.items()
            if rank >= threshold
        )

        if rank_sum < min_counts:
            continue

        new_counts[taxon] = rank_sum

    return new_counts


def basic_stats(array, sep):
    """
    Returns formatted basic statistics for contig lengths
    """
    amedian = numpy.median(array)
    amean = numpy.mean(array)
    amax = numpy.max(array)
    amin = numpy.min(array)
    return sep.join(str(value) for value in (amedian, amean, amax, amin))


def write_fasta_summary(file_handle, seq_lengths, seq_lengths_filt, sep='\t'):
    """
    Write summary file for assembly

    :param file_handle: file handle for output
    :param array seq_lengths: array for sequence lengths
    :param array seq_lengths_filt: array for sequence lengths of annotated
        contigs
    :param str sep: string used as column separator
    """
    n50 = calc_n50(seq_lengths)
    file_handle.write("Assembly N50:{0}{1}\n\n".format(sep, n50))
    header = "{0}\n".format(sep.join(("", "median", "mean", "max", "min")))
    file_handle.write(header)
    file_handle.write("assembly{0}{1}\n".format(sep,
                      basic_stats(seq_lengths, sep)))
    file_handle.write("annotated contigs{0}{1}\n".format(sep,
                      basic_stats(seq_lengths_filt, sep)))



"""
Module dealing with BAM/SAM files
"""

import logging
import itertools

from . import DependencyError

try:
    import pandas
    import numpy
    import pysam
except ImportError:
    raise DependencyError('numpy, pysam, pandas')

LOG = logging.getLogger(__name__)

try:
    from progress.bar import Bar
except ImportError:
    LOG.debug("Could not import progress module")
    Bar = None


def get_region_coverage(bam_file, seq_id, feat_from, feat_to):
    """
    Return coverage for an annotation.

    .. note::

        feat_from and feat_to are 1-based indexes

    :param Samfile bam_file: instance of :class:`pysam.Samfile`
    :param str seq_id: sequence id
    :param int feat_from: start position of feature
    :param int feat_to: end position of feature

    :return int: coverage array for the annotation
    """
    norm_start, norm_end = feat_from - 1, feat_to - 1
    iterator = bam_file.pileup(
        reference=seq_id,
        start=norm_start,
        end=norm_end,
        truncate=True
    )

    coverage = pandas.Series(
        dict(
            (pileup_proxy.pos, pileup_proxy.n)
            for pileup_proxy in iterator
        ),
        dtype=numpy.int,
    )

    return coverage


def covered_annotation_bp(files, annotations, min_cov=1, progress=False):
    """
    .. versionadded:: 0.1.14

    Returns the number of base pairs covered of annotations over multiple
    samples.

    Arguments:
        files (iterable): an iterable that returns the alignment file names
        annotations (iterable): an iterable that returns annotations
        min_cov (int): minumum coverage for a base to counted
        progress (bool): if *True*, a progress bar is used

    Returns:
        dict: a dictionary whose keys are the uid and the values the number of
        bases that are covered by reads among all samples

    """
    annotations = [
        (annotation.uid, annotation.seq_id, annotation.start, annotation.end)
        for annotation in annotations
    ]

    covered = {}

    for file_name in files:
        # pysam 0.8.1 changed Samfile to AlignmentFile
        alg_file = pysam.AlignmentFile(file_name, 'rb')

        if progress and (Bar is not None):
            bar = Bar(
                file_name,
                max=len(annotations),
                suffix='%(percent)d%% - %(eta_td)s'
            )
            bar.width = 64
        else:
            bar = None

        for uid, seq_id, start, end in annotations:
            cov = get_region_coverage(alg_file, seq_id, start, end)

            try:
                covered[uid] = cov.add(covered[uid], fill_value=0)
            except KeyError:
                covered[uid] = cov

            if bar is not None:
                bar.next()

        if bar is not None:
            bar.finish()

    return dict(
        (uid, len(cov[cov >= min_cov])) for uid, cov in covered.iteritems()
    )


def add_coverage_info(annotations, bam_files, samples, attr_suff='_cov'):
    """
    Adds coverage information to annotations, using BAM files.

    The coverage information is added for each sample as a 'sample_cov' and the
    total coverage as as 'cov' attribute in the annotations.

    .. note::

        The bam_files and sample variables must have the same order

    :param iterable annotations: iterable of annotations
    :param iterable bam_files: iterable of :class:`pysam.Samfile` instances
    :param iterable sample: names of the samples for the BAM files
    """
    LOG.info("Adding coverage info for %d annotations", len(annotations))
    for bam_file, sample in zip(bam_files, samples):
        LOG.info("Sample %s, file %s", sample, bam_file.filename)

    tot_coverage = {}

    for bam_file, sample in itertools.izip(bam_files, samples):
        LOG.info("Adding coverage for sample %s", sample)
        for index, annotation in enumerate(annotations):
            sample_coverage = get_region_coverage(
                bam_file,
                annotation.seq_id,
                annotation.start,
                annotation.end
            )
            # adds the results of the coverage to the total coverage
            # uses the add method of a Series to make sure that possible
            # nan values are filled with 0 before the sum
            try:
                tot_coverage[annotation.uid] = sample_coverage.add(
                    tot_coverage[annotation.uid],
                    fill_value=0
                )
            except KeyError:
                tot_coverage[annotation.uid] = sample_coverage

            cov_mean = sample_coverage.mean()

            annotation.set_attr(
                sample + attr_suff,
                0 if numpy.isnan(cov_mean) else int(cov_mean)
            )
            if (index + 1) % 1000 == 0:
                LOG.debug(
                    'Added coverage for (%d/%d) annotations',
                    index + 1,
                    len(annotations)
                )
    for annotation in annotations:
        cov_mean = tot_coverage[annotation.uid].mean()
        annotation.set_attr(
            'cov',
            0 if numpy.isnan(cov_mean) else int(cov_mean)
        )


def read_samtools_depth(file_handle, num_seqs=10000):
    """
    .. versionadded:: 0.3.0

    Reads a samtools *depth* file, returning a generator that yields the
    array of each base coverage on a per-sequence base.

    .. note::

        The information on position is not used, to use numpy and save memory.
        samtools *depth* should be called with the `-aa` option::

             `samtools depth -aa bamfile`

        This options will output both base position with 0 coverage and
        sequneces with no aligned reads

    Arguments:
        file_handle (file): file handle of the coverage file
        num_seqs (int): number of sequence that fires a log message

    Yields:
        tuple: the first element is the sequence identifier and the second one
        is the *numpy* array with the positions
    """
    curr_key = ''
    curr_cov = []
    LOG.info(
        'Reading coverage from file (%s)',
        getattr(file_handle, 'name', repr(file_handle))
    )
    line_no = 0
    for line in file_handle:
        name, pos, cov = line.strip().split('\t')
        cov = int(cov)
        if curr_key == name:
                curr_cov.append(cov)
        else:
            if curr_key == '':
                curr_cov.append(cov)
                curr_key = name
            else:
                line_no += 1
                if line_no % num_seqs == 0:
                    LOG.info('Read %d sequence coverage', line_no)
                yield curr_key, numpy.array(curr_cov)
                curr_key = name
                curr_cov = [cov]
    else:
        yield curr_key, numpy.array(curr_cov)

    LOG.info('Read a total of %d sequence coverage', line_no + 1)


class SamtoolsDepth(object):
    """
    .. versionadded:: 0.3.0

    A class used to cache the results of :func:`read_samtools_depth`, while
    reading only the necessary data from a`samtools depth -aa` file.
    """
    file_handle = None
    data = None

    def __init__(self, file_handle, num_seqs=10**4):
        """
        Arguments:
            file_handle (file): the file handle to pass to
                :func:`read_samtools_depth`
            num_seqs (int): number of sequence that fires a log message
        """
        self.file_handle = read_samtools_depth(file_handle, num_seqs=num_seqs)
        self.data = {}

    def region_coverage(self, seq_id, start, end):
        """
        Returns the mean coverage of a region. The *start* and *end* parameters
        are expected to be 1-based coordinates, like the correspondent
        attributes in :class:`mgkit.io.gff.Annotation` or
        :class:`mgkit.io.gff.GenomicRange`.

        If the sequence for which the coverage is requested is not found, the
        *depth* file is read (and cached) until it is found.

        Arguments:
            seq_id (str): sequence for which to return mean coverage
            start (int): start of the region
            end (int): end of the region

        Returns:
            float: mean coverage of the requested region
        """
        try:
            cov = self.data[seq_id][start-1:end]
        except KeyError:
            for key, value in self.file_handle:
                self.data[key] = value
                # if the key is the one requested, the loop is stopped and and
                # the value is kept
                if key == seq_id:
                    cov = value[start-1:end]
                    break

        return cov.mean()



"""
Dictionary utils

"""
import numpy


def combine_dict(keydict, valuedict):
    """
    Combine two dictionaries when the values of keydict are iterables. The
    combined dictionary has the same keys as keydict and the its values are
    sets containing all the values associated to keydict values in valuedict.

    .. digraph:: keydict
        :alt: key1 -> [v1, v2, .., vN]

        "keydict" -> "key1" -> "[v1, v2, .., vN]";

    .. digraph:: valuedict
        :alt: v1 -> [u1, u2, .., uN]
              v2 -> [t1, t2, .., tN]

        "valuedict" -> "v1" -> "[u1, u2, .., uN]";
        "valuedict" -> "v2" -> "[t1, t2, .., tN]";

    Resulting dictionary will be

    .. digraph:: combined
        :alt: key1->{u1, u2, .., uN}

        "combined" -> "key1" -> "{u1, u2, .., uN, t1, t2, .., tN}";

    :param dict keydict: dictionary whose keys are the same as the returned
        dictionary
    :param dict valuedict: dictionary whose values are the same as the returned
        dictionary

    :return dict: combined dictionary
    """

    comb_dict = dict((key, set()) for key in keydict)

    for key, values in keydict.items():
        for value in values:
            try:
                comb_dict[key].update(valuedict[value])
            except KeyError:
                # in case a value isn't in valuedict keys, silently pass
                pass

    return comb_dict


def combine_dict_one_value(keydict, valuedict):
    """
    Combine two dictionaries by the value of the keydict is used as a key in
    valuedict and the resulting dictionary is composed of keydict keys and
    valuedict values.

    Same as :func:`comb_dict`, but each value in keydict is a single element
    that is key in valuedict.

    :param dict keydict: dictionary whose keys are the same as the returned
        dictionary
    :param dict valuedict: dictionary whose values are the same as the returned
        dictionary

    :return dict: combined dictionary
    """

    comb_dict = {}

    for key, value in keydict.items():
        comb_dict[key] = valuedict[value]

    return comb_dict


def link_ids(id_map, black_list=None):
    """
    Given a dictionary whose values (iterables) can be linked back to other
    keys, it returns a dictionary in which the keys are the original keys and
    the values are sets of keys to which they can be linked.

    .. digraph:: input
        :alt: key1->[v1, v2]
              key2->[v3, v4]
              key3->[v2, v4]

        "id_map" -> "key1" -> "[v1, v2]";
        "id_map" -> "key2" -> "[v3, v4]";
        "id_map" -> "key3" -> "[v2, v4]";

    Becomes:

    .. digraph:: output
        :alt: key1->[key1, key3]
              key2->[key3]
              key3->[key1, key2]

        "linked" -> "key1" -> "[key1, key3]"
        "linked" -> "key2" -> "[key3]"
        "linked" -> "key3" -> "[key1, key2]"

    :param dict id_map: dictionary of keys to link
    :param iterable black_list: iterable of values to skip in making the links

    :return dict: linked dictionary
    """
    id_to_id = {}

    for s_id, s_cps in id_map.iteritems():
        id_links = set()
        for e_id, e_cps in id_map.iteritems():
            if e_id == s_id:
                continue
            for s_id2 in s_cps:
                if black_list is not None:
                    if s_id2 in black_list:
                        continue
                if s_id2 in e_cps:
                    id_links.add(e_id)
                    break
        id_to_id[s_id] = id_links
    return id_to_id


def reverse_mapping(map_dict):
    """
    Given a dictionary in the form: key->[v1, v2, .., vN], returns a dictionary
    in the form: v1->[key1, key2, .., keyN]

    :param dict map_dict: dictionary to reverse

    :return dict: reversed dictionary
    """
    rev_map = {}

    for key, value_ids in map_dict.iteritems():
        for value_id in value_ids:
            try:
                rev_map[value_id].add(key)
            except KeyError:
                rev_map[value_id] = set([key])

    return rev_map


def find_id_in_dict(s_id, s_dict):
    """
    Finds a value 's_id' in a dictionary in which the values are iterables.
    Returns a list of keys that contain the value.

    :param dict s_id: element to look for in the dictionary's values
    :param object d: dictionary to search in

    :return list: list of keys in which d was found
    """
    f_list = []
    for k_id, v_ids in s_dict.items():
        if s_id in v_ids:
            f_list.append(k_id)
    return f_list


def split_dictionary_by_value(value_dict, threshold, aggr_func=numpy.median,
                              key_filter=None):
    """
    Splits a dictionary, whose values are iterables, based on a threshold:

        * one in which the result of aggr_func is lower than the threshold
          (first)
        * one in which the result of aggr_func is equal or greater than the
          threshold (second)

    :param dict valuedict: dictionary to be splitted
    :param number threshold: must be comparable to threshold
    :param func aggr_func: function used to aggregate the dictionary values
    :param iterable key_filter: if specified, only these key will be in the
        resulting dictionary

    :return: two dictionaries
    """
    lower_dict = {}
    higher_dict = {}

    if key_filter is None:
        key_filter = value_dict.iterkeys()

    for key in key_filter:
        values = value_dict[key]
        if aggr_func(values) < threshold:
            lower_dict[key] = values
        else:
            higher_dict[key] = values

    return lower_dict, higher_dict


def apply_func_to_values(dictionary, func):
    """
    .. versionadded:: 0.1.12

    Assuming a dictionary whose values are iterables, *func* is applied to each
    element of the iterable, retuning a *set* of all transformed elements.

    Arguments:
        dictionary (dict): dictionary whose values are iterables
        func (func): function to apply to the dictionary values

    Returns:
        dict: dictionary with transformed values
    """
    return dict(
        (key, set(func(value) for value in values))
        for key, values in dictionary.iteritems()
    )


def filter_ratios_by_numbers(ratios, min_num):
    """
    Returns from a dictionary only the items for which the length of the
    iterables that is the value of the item, is equal or greater of min_num.

    :param dict ratios: dictionary key->list
    :param int min_num: minimum number of elements in the value iterable

    :return dict: filtered dictionary
    """
    return dict(
        (key, values) for key, values in ratios.iteritems()
        if len(values) >= min_num
    )


def filter_nan(ratios):
    """
    Returns a dictionary with the NaN values taken out
    """

    return dict(
        (key, [ratio for ratio in ratios[key] if not numpy.isnan(ratio)])
        for key in ratios
    )


class cache_dict_file(object):
    """
    .. versionadded:: 0.3.0

    Used to cache the result of a function that yields a tuple (key and value).
    If the value is found in the internal dictionary (as the class behave), the
    correspondent value is returned, otherwise the iterator is advanced until
    the key is found.

    Example:
        >>> from mgkit.io.blast import parse_accession_taxa_table
        >>> i = parse_accession_taxa_table('nucl_gb.accession2taxid.gz', key=0)
        >>> d = cache_dict_file(i)
        >>> d['AH001684']
        4400
    """
    _iterator = None
    _dict = None

    def __init__(self, iterator, skip_lines=0):
        """
        Arguments:
            iterator (iter): iterator used in building the dictionary
            skip_lines (int): how many iterations to skip at the start
        """
        for index in xrange(skip_lines):
            next(iterator)
        self._iterator = iterator
        self._dict = {}

    def __getitem__(self, key):
        try:
            value = self._dict[key]
        except KeyError:
            while True:
                nkey, nvalue = self.next()
                if nkey == key:
                    value = nvalue
                    break
        return value

    def next(self):
        try:
            key, value = next(self._iterator)
        except StopIteration:
            raise KeyError
        self._dict[key] = value
        return key, value



"""
.. deprecated:: 0.2.3

Functions related to handling :mod:`pandas` data structures
"""


def filter_dataframe_by_count(dataframe, min_count, no_empty_rows=True):
    """
    Filter the number of column of a DataFrame based on a minimum amount of
    non-NA counts and removes all rows that have no non-NA values if
    no_empty_rows is True (default)

    :param DataFrame dataframe: :class:`pandas.DataFrame` instance
    :param int min_count: minimum number of non-NA counts for columns
    :param bool no_empty_rows: if True, removes all empty rows

    :return: :class:`pandas.DataFrame` instance
    """

    cols = dataframe.ix[:, dataframe.count(axis=0) > min_count].columns
    dataframe = dataframe.ix[:, cols]
    if no_empty_rows:
        rows = dataframe[dataframe.count(axis=1) > 0].index
    else:
        rows = dataframe.index

    return dataframe.ix[rows]



"""
Script to convert HMMER results files (domain table) to a GFF file, the name of
the profiles are expected to be now in the form
*GENEID_TAXONID_TAXON-NAME(-nr)* by default, but any other profile name is
accepted.

The profiles tested are those made from Kegg Orthologs, from the
`download_profiles` script. If the `--no-custom-profiles` options is used,
the script can be used with any profile name. The profile name will be used
for `gene_id`, `taxon_id` and `taxon_name` in the GFF file.

.. note::

    for GENEID, old documentation points to KOID, it is the same

.. warning::

    The compatibility with old data has been **removed**, meaning that old
    experiments must use the scripts from those versions. It is possible to use
    multiple environments, with `virtualenv` for this purpose. An examples is
    given in :ref:`install-ref`.

Changes
*******

.. versionchanged:: 0.1.15
    adapted to new GFF module and specs

.. versionchanged:: 0.2.1
    added options to customise output and filters and old restrictions

"""

import sys
import logging
import argparse
from mgkit import logger
from mgkit.io import gff
from mgkit.io import fasta
from . import utils

LOG = logging.getLogger(__name__)


def set_parser():
    """
    Setup command line options
    """
    parser = argparse.ArgumentParser(
        description='Convert HMMER data to GFF file',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    group = parser.add_argument_group('File options')
    group.add_argument(
        'aa_file',
        type=argparse.FileType('r'),
        help="Fasta file containing contigs translated to aa (used by HMMER)"
    )
    group.add_argument(
        'hmmer_file',
        nargs='?',
        type=argparse.FileType('r'),
        default='-'
    )
    group.add_argument(
        '-o',
        '--output-file',
        nargs='?',
        type=argparse.FileType('w'), default=sys.stdout
    )

    group = parser.add_argument_group('Filters')
    group.add_argument(
        '-t',
        '--discard',
        action='store',
        type=float,
        default=0.05,
        help='Evalue over which an hit will be discarded'
    )
    group.add_argument(
        '-d',
        '--disable-evalue',
        action='store_true',
        default=False,
        help='Disable Evalue filter'
    )

    group = parser.add_argument_group('GFF')
    group.add_argument(
        '-c',
        '--no-custom-profiles',
        action='store_false',
        default=True,
        help='Profiles names are not in the custom format'
    )
    group.add_argument(
        '-db',
        '--database',
        action='store',
        default='CUSTOM',
        help='Database from which the profiles are generated " +" (e.g. PFAM)'
    )
    group.add_argument(
        '-f',
        '--feature-type',
        action='store',
        default='gene',
        help='Type of feature (e.g. gene)'
    )

    group = parser.add_argument_group('Misc')

    utils.add_basic_options(parser, manual=__doc__)

    return parser


def get_aa_data(f_handle):
    """
    Load aminoacid seuqnces used by HMMER.
    """
    # LOG.info('Loading aa data from file %s', f_handle.name)

    aa_seqs = dict((name, seq) for name, seq in fasta.load_fasta(f_handle))

    return aa_seqs


def parse_domain_table_contigs(options):
    """
    Parse the HMMER result file
    """
    aa_seqs = get_aa_data(options.aa_file)

    LOG.info('Parsing HMMER data from file %s', options.hmmer_file.name)
    LOG.info('Writing GFF data to file %s', options.output_file.name)

    count_dsc = 0
    count_tot = 0
    count_skp = 0

    for idx, line in enumerate(options.hmmer_file):

        if line.startswith('#'):
            continue
        if idx % 10000 == 0:
            LOG.info("Line number: %d", idx)

        count_tot += 1

        try:
            annotation = gff.from_hmmer(
                line,
                aa_seqs,
                feat_type=options.feature_type,
                db=options.database,
                custom_profiles=options.no_custom_profiles
            )
        except ZeroDivisionError:
            LOG.error(
                "Skipping line %d because of an error in the calculations",
                idx + 1
            )
            count_skp += 1
            continue

        # if disable_evalue is True, skips filter
        if not options.disable_evalue:
            if annotation.score > options.discard:
                count_dsc += 1
                continue

        annotation.to_file(options.output_file)

    LOG.info(
        "Read %d lines, discarded %d, skipped %d",
        count_tot,
        count_dsc,
        count_skp
    )


def main():
    """
    Main loop
    """
    options = set_parser().parse_args()
    logger.config_log(options.verbose)

    parse_domain_table_contigs(
        options
    )

if __name__ == '__main__':
    main()



"""
Package used to contain code related to I/O operations
"""

import gzip
import bz2

try:
    import lzma
except ImportError:
    lzma = None


class UnsupportedFormat(IOError):
    "Raised if the a file can't be opened with the correct module"
    pass


def open_file(file_name, mode='r'):
    """
    .. versionadded:: 0.1.12

    Opens a file using the extension as a guide to which module to use.

    Arguments:
        file_name (str): file name
        mode (str): mode used to open the file

    Returns:
        file: file handle

    Raises:
        UnsupportedFormat: if the module to open the file is not available

    """
    if file_name.endswith('.gz'):
        file_handle = gzip.GzipFile(file_name, mode)
    elif file_name.endswith('.bz2'):
        file_handle = bz2.BZ2File(file_name, mode)
    elif file_name.endswith('.xz'):
        if lzma:
            raise UnsupportedFormat("Cannot import lzma module")
        else:
            file_handle = lzma.LZMAFile(file_name, mode)
    else:
        file_handle = open(file_name, mode)

    return file_handle


def compressed_handle(file_handle):
    """
    .. versionadded:: 0.1.13

    Tries to wrap a file handle in the appropriate compressed file class.

    Arguments:
        file_handle (str): file handle

    Returns:
        file: the same file handle if no suitable compressed file class is
        found or the new file_handle which supports the compression

    Raises:
        UnsupportedFormat: if the module to open the file is not available

    """
    if file_handle.name.endswith('.gz'):
        file_handle = gzip.GzipFile(fileobj=file_handle, mode='rb')
    elif file_handle.name.endswith('.xz'):
        if lzma:
            raise UnsupportedFormat("Cannot import lzma module")
        else:
            file_handle = lzma.LZMAFile(file_handle)

    return file_handle


def split_write(records, name_mask, write_func, num_files=2):
    """
    .. versionadded:: 0.1.13

    Splits the writing of a number of records in a series of files. The
    `name_mask` is used as template for the file names. A string like
    "split-files-{0}" can be specified and the function applies format with the
    index of the pieces.

    Arguments:
        records (iterable): an iterable that returns a object to be saved
        name_mask (str): a string used as template for the output file names
            on which the function applies :func:`string.format`
        write_func (func): a function that is called to write to the files. It
            needs to accept a file handles as first argument and the record
            returned by `records` as the second argument
        num_files (int): the number of files to split the records
    """
    out_handles = [open_file(name_mask.format(x), 'w') for x in xrange(num_files)]

    for index, record in enumerate(records):
        out_handle = out_handles[index % num_files]
        write_func(out_handle, record)



"""
Access some R functions. requires :mod:`rpy2`
"""

# to get python 3.x behavior on relative imports - it's absolute unless
# specified by a dot - otherwise current dir module is loaded
from __future__ import absolute_import
from .. import DependencyError

try:
    import rpy2.robjects as robjects
except ImportError:
    raise DependencyError('rpy')

try:
    import pandas
except ImportError:
    raise DependencyError('pandas')


def correct_pvalues(pvalues, method='BH'):
    """
    Correct pvalues using R.

    :param pandas.Series pvalues: :class:`pandas.Series` instance containing
        the pvalues
    :param str method: string passed to the R function to specify the
        correction method

    :return: a :class:`pandas.Series` with the same index as pvalues
    """

    padj = robjects.r('p.adjust')

    corr_pvalues = padj(robjects.FloatVector(pvalues), method)
    corr_pvalues = pandas.Series(corr_pvalues, index=pvalues.index)

    return corr_pvalues
